##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""RML ``document`` element
"""
import io
import logging

import reportlab.pdfgen.canvas
import zope.interface
from reportlab.lib import colors
from reportlab.lib import fonts
from reportlab.pdfbase import cidfonts
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase import ttfonts
from reportlab.platypus import tableofcontents
from reportlab.platypus.doctemplate import IndexingFlowable

from z3c.rml import attr
from z3c.rml import canvas
from z3c.rml import directive
from z3c.rml import doclogic  # noqa: F401 imported but unused
from z3c.rml import interfaces
from z3c.rml import list  # noqa: F401 imported but unused
from z3c.rml import occurence
from z3c.rml import pdfinclude  # noqa: F401 imported but unused
from z3c.rml import special
from z3c.rml import storyplace  # noqa: F401 imported but unused
from z3c.rml import stylesheet
from z3c.rml import template


LOGGER_NAME = 'z3c.rml.render'


class IRegisterType1Face(interfaces.IRMLDirectiveSignature):
    """Register a new Type 1 font face."""

    afmFile = attr.File(
        title='AFM File',
        description='Path to AFM file used to register the Type 1 face.',
        doNotOpen=True,
        required=True)

    pfbFile = attr.File(
        title='PFB File',
        description='Path to PFB file used to register the Type 1 face.',
        doNotOpen=True,
        required=True)


class RegisterType1Face(directive.RMLDirective):
    signature = IRegisterType1Face

    def process(self):
        args = self.getAttributeValues(valuesOnly=True)
        face = pdfmetrics.EmbeddedType1Face(*args)
        pdfmetrics.registerTypeFace(face)


class IRegisterFont(interfaces.IRMLDirectiveSignature):
    """Register a new font based on a face and encoding."""

    name = attr.Text(
        title='Name',
        description=('The name under which the font can be used in style '
                     'declarations or other parameters that lookup a font.'),
        required=True)

    faceName = attr.Text(
        title='Face Name',
        description=('The name of the face the font uses. The face has to '
                     'be previously registered.'),
        required=True)

    encName = attr.Text(
        title='Encoding Name',
        description=('The name of the encdoing to be used.'),
        required=True)


class RegisterFont(directive.RMLDirective):
    signature = IRegisterFont

    def process(self):
        args = self.getAttributeValues(valuesOnly=True)
        font = pdfmetrics.Font(*args)
        pdfmetrics.registerFont(font)


class IAddMapping(interfaces.IRMLDirectiveSignature):
    """Map various styles(bold, italic) of a font name to the actual ps fonts
    used."""

    faceName = attr.Text(
        title='Name',
        description=('The name of the font to be mapped'),
        required=True)

    bold = attr.Integer(
        title='Bold',
        description=('Bold'),
        required=True)

    italic = attr.Integer(
        title='Italic',
        description=('Italic'),
        required=True)

    psName = attr.Text(
        title='psName',
        description=('Actual font name mapped'),
        required=True)


class AddMapping(directive.RMLDirective):
    signature = IAddMapping

    def process(self):
        args = self.getAttributeValues(valuesOnly=True)
        fonts.addMapping(*args)


class IRegisterTTFont(interfaces.IRMLDirectiveSignature):
    """Register a new TrueType font given the TT file and face name."""

    faceName = attr.Text(
        title='Face Name',
        description=('The name of the face the font uses. The face has to '
                     'be previously registered.'),
        required=True)

    fileName = attr.File(
        title='File Name',
        description='File path of the of the TrueType font.',
        doNotOpen=True,
        doNotModify=True,
        required=True)


class RegisterTTFont(directive.RMLDirective):
    signature = IRegisterTTFont

    def process(self):
        args = self.getAttributeValues(valuesOnly=True)
        font = ttfonts.TTFont(*args)
        pdfmetrics.registerFont(font)


class IRegisterCidFont(interfaces.IRMLDirectiveSignature):
    """Register a new CID font given the face name."""

    faceName = attr.Text(
        title='Face Name',
        description=('The name of the face the font uses. The face has to '
                     'be previously registered.'),
        required=True)

    encName = attr.Text(
        title='Encoding Name',
        description=('The name of the encoding to use for the font.'),
        required=False)


class RegisterCidFont(directive.RMLDirective):
    signature = IRegisterCidFont
    attrMapping = {'faceName': 'face', 'encName': 'encoding'}

    def process(self):
        args = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        if 'encoding' in args:
            font = cidfonts.CIDFont(**args)
        else:
            font = cidfonts.UnicodeCIDFont(**args)
        pdfmetrics.registerFont(font)


class IRegisterFontFamily(interfaces.IRMLDirectiveSignature):
    """Register a new font family."""

    name = attr.Text(
        title='Name',
        description=('The name of the font family.'),
        required=True)

    normal = attr.Text(
        title='Normal Font Name',
        description=('The name of the normal font variant.'),
        required=False)

    bold = attr.Text(
        title='Bold Font Name',
        description=('The name of the bold font variant.'),
        required=False)

    italic = attr.Text(
        title='Italic Font Name',
        description=('The name of the italic font variant.'),
        required=False)

    boldItalic = attr.Text(
        title='Bold/Italic Font Name',
        description=('The name of the bold/italic font variant.'),
        required=True)


class RegisterFontFamily(directive.RMLDirective):
    signature = IRegisterFontFamily
    attrMapping = {'name': 'family'}

    def process(self):
        args = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        pdfmetrics.registerFontFamily(**args)


class IColorDefinition(interfaces.IRMLDirectiveSignature):
    """Define a new color and give it a name to be known under."""

    id = attr.Text(
        title='Id',
        description=('The id/name the color will be available under.'),
        required=True)

    RGB = attr.Color(
        title='RGB Color',
        description=('The color value that is represented.'),
        required=False)

    CMYK = attr.Color(
        title='CMYK Color',
        description=('The color value that is represented.'),
        required=False)

    value = attr.Color(
        title='Color',
        description=('The color value that is represented.'),
        required=False)

    spotName = attr.Text(
        title='Spot Name',
        description=('The Spot Name of the CMYK color.'),
        required=False)

    density = attr.Float(
        title='Density',
        description=('The color density of the CMYK color.'),
        min=0.0,
        max=1.0,
        required=False)

    knockout = attr.Text(
        title='Knockout',
        description=('The knockout of the CMYK color.'),
        required=False)

    alpha = attr.Float(
        title='Alpha',
        description=('The alpha channel of the color.'),
        min=0.0,
        max=1.0,
        required=False)


class ColorDefinition(directive.RMLDirective):
    signature = IColorDefinition

    def process(self):
        kwargs = dict(self.getAttributeValues())
        id = kwargs.pop('id')
        for attrName in ('RGB', 'CMYK', 'value'):
            color = kwargs.pop(attrName, None)
            if color is not None:
                # CMYK has additional attributes.
                for name, value in kwargs.items():
                    setattr(color, name, value)
                manager = attr.getManager(self)
                manager.colors[id] = color
                return
        raise ValueError('At least one color definition must be specified.')


# Initialize also supports the <color> tag.
stylesheet.Initialize.factories['color'] = ColorDefinition
stylesheet.IInitialize.setTaggedValue(
    'directives',
    stylesheet.IInitialize.getTaggedValue('directives') +
    (occurence.ZeroOrMore('color', IColorDefinition),)
)


class IStartIndex(interfaces.IRMLDirectiveSignature):
    """Start a new index."""

    name = attr.Text(
        title='Name',
        description='The name of the index.',
        default='index',
        required=True)

    offset = attr.Integer(
        title='Offset',
        description='The counting offset.',
        min=0,
        required=False)

    format = attr.Choice(
        title='Format',
        description=('The format the index is going to use.'),
        choices=interfaces.LIST_FORMATS,
        required=False)


class StartIndex(directive.RMLDirective):
    signature = IStartIndex

    def process(self):
        kwargs = dict(self.getAttributeValues())
        name = kwargs['name']
        manager = attr.getManager(self)
        manager.indexes[name] = tableofcontents.SimpleIndex(**kwargs)


class ICropMarks(interfaces.IRMLDirectiveSignature):
    """Crop Marks specification"""

    name = attr.Text(
        title='Name',
        description='The name of the index.',
        default='index',
        required=True)

    borderWidth = attr.Measurement(
        title='Border Width',
        description='The width of the crop mark border.',
        required=False)

    markColor = attr.Color(
        title='Mark Color',
        description='The color of the crop marks.',
        required=False)

    markWidth = attr.Measurement(
        title='Mark Width',
        description='The line width of the actual crop marks.',
        required=False)

    markLength = attr.Measurement(
        title='Mark Length',
        description='The length of the actual crop marks.',
        required=False)

    markLast = attr.Boolean(
        title='Mark Last',
        description='If set, marks are drawn after the content is rendered.',
        required=False)

    bleedWidth = attr.Measurement(
        title='Bleed Width',
        description=('The width of the page bleed.'),
        required=False)


class CropMarksProperties:
    borderWidth = 36
    markWidth = 0.5
    markColor = colors.toColor('green')
    markLength = 18
    markLast = True
    bleedWidth = 0


class CropMarks(directive.RMLDirective):
    signature = ICropMarks

    def process(self):
        cmp = CropMarksProperties()
        for name, value in self.getAttributeValues():
            setattr(cmp, name, value)
        self.parent.parent.cropMarks = cmp


class ILogConfig(interfaces.IRMLDirectiveSignature):
    """Configure the render logger."""

    level = attr.Choice(
        title='Level',
        description='The default log level.',
        choices=interfaces.LOG_LEVELS,
        doLower=False,
        required=False)

    format = attr.Text(
        title='Format',
        description='The format of the log messages.',
        required=False)

    filename = attr.File(
        title='File Name',
        description='The path to the file that is being logged.',
        doNotOpen=True,
        required=True)

    filemode = attr.Choice(
        title='File Mode',
        description='The mode to open the file in.',
        choices={'WRITE': 'w', 'APPEND': 'a'},
        default='a',
        required=False)

    datefmt = attr.Text(
        title='Date Format',
        description='The format of the log message date.',
        required=False)


class LogConfig(directive.RMLDirective):
    signature = ILogConfig

    def process(self):
        args = dict(self.getAttributeValues())
        logger = logging.getLogger(LOGGER_NAME)
        handler = logging.FileHandler(args['filename'][8:], args['filemode'])
        formatter = logging.Formatter(
            args.get('format'), args.get('datefmt'))
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        if 'level' in args:
            logger.setLevel(args['level'])
        self.parent.parent.logger = logger


class IDocInit(interfaces.IRMLDirectiveSignature):
    occurence.containing(
        occurence.ZeroOrMore('color', IColorDefinition),
        occurence.ZeroOrMore('name', special.IName),
        occurence.ZeroOrMore('registerType1Face', IRegisterType1Face),
        occurence.ZeroOrMore('registerFont', IRegisterFont),
        occurence.ZeroOrMore('registerCidFont', IRegisterCidFont),
        occurence.ZeroOrMore('registerTTFont', IRegisterTTFont),
        occurence.ZeroOrMore('registerFontFamily', IRegisterFontFamily),
        occurence.ZeroOrMore('addMapping', IAddMapping),
        occurence.ZeroOrMore('logConfig', ILogConfig),
        occurence.ZeroOrMore('cropMarks', ICropMarks),
        occurence.ZeroOrMore('startIndex', IStartIndex),
    )

    pageMode = attr.Choice(
        title='Page Mode',
        description=('The page mode in which the document is opened in '
                     'the viewer.'),
        choices=('UseNone', 'UseOutlines', 'UseThumbs', 'FullScreen'),
        required=False)

    pageLayout = attr.Choice(
        title='Page Layout',
        description=('The layout in which the pages are displayed in '
                     'the viewer.'),
        choices=('SinglePage', 'OneColumn', 'TwoColumnLeft', 'TwoColumnRight'),
        required=False)

    useCropMarks = attr.Boolean(
        title='Use Crop Marks',
        description='A flag when set shows crop marks on the page.',
        required=False)

    hideToolbar = attr.TextBoolean(
        title='Hide Toolbar',
        description=('A flag indicating that the toolbar is hidden in '
                     'the viewer.'),
        required=False)

    hideMenubar = attr.TextBoolean(
        title='Hide Menubar',
        description=('A flag indicating that the menubar is hidden in '
                     'the viewer.'),
        required=False)

    hideWindowUI = attr.TextBoolean(
        title='Hide Window UI',
        description=('A flag indicating that the window UI is hidden in '
                     'the viewer.'),
        required=False)

    fitWindow = attr.TextBoolean(
        title='Fit Window',
        description='A flag indicating that the page fits in the viewer.',
        required=False)

    centerWindow = attr.TextBoolean(
        title='Center Window',
        description=('A flag indicating that the page fits is centered '
                     'in the viewer.'),
        required=False)

    displayDocTitle = attr.TextBoolean(
        title='Display Doc Title',
        description=('A flag indicating that the document title is displayed '
                     'in the viewer.'),
        required=False)

    nonFullScreenPageMode = attr.Choice(
        title='Non-Full-Screen Page Mode',
        description=('Non-Full-Screen page mode in the viewer.'),
        choices=('UseNone', 'UseOutlines', 'UseThumbs', 'UseOC'),
        required=False)

    direction = attr.Choice(
        title='Text Direction',
        description=('The text direction of the PDF.'),
        choices=('L2R', 'R2L'),
        required=False)

    viewArea = attr.Choice(
        title='View Area',
        description=('View Area setting used in the viewer.'),
        choices=('MediaBox', 'CropBox', 'BleedBox', 'TrimBox', 'ArtBox'),
        required=False)

    viewClip = attr.Choice(
        title='View Clip',
        description=('View Clip setting used in the viewer.'),
        choices=('MediaBox', 'CropBox', 'BleedBox', 'TrimBox', 'ArtBox'),
        required=False)

    printArea = attr.Choice(
        title='Print Area',
        description=('Print Area setting used in the viewer.'),
        choices=('MediaBox', 'CropBox', 'BleedBox', 'TrimBox', 'ArtBox'),
        required=False)

    printClip = attr.Choice(
        title='Print Clip',
        description=('Print Clip setting used in the viewer.'),
        choices=('MediaBox', 'CropBox', 'BleedBox', 'TrimBox', 'ArtBox'),
        required=False)

    printScaling = attr.Choice(
        title='Print Scaling',
        description=('The print scaling mode in which the document is opened '
                     'in the viewer.'),
        choices=('None', 'AppDefault'),
        required=False)


class DocInit(directive.RMLDirective):
    signature = IDocInit
    factories = {
        'name': special.Name,
        'color': ColorDefinition,
        'registerType1Face': RegisterType1Face,
        'registerFont': RegisterFont,
        'registerTTFont': RegisterTTFont,
        'registerCidFont': RegisterCidFont,
        'registerFontFamily': RegisterFontFamily,
        'addMapping': AddMapping,
        'logConfig': LogConfig,
        'cropMarks': CropMarks,
        'startIndex': StartIndex,
    }

    viewerOptions = {
        option[0].lower() + option[1:]: option
        for option in ['HideToolbar', 'HideMenubar', 'HideWindowUI',
                       'FitWindow', 'CenterWindow', 'DisplayDocTitle',
                       'NonFullScreenPageMode', 'Direction', 'ViewArea',
                       'ViewClip', 'PrintArea', 'PrintClip', 'PrintScaling']}

    def process(self):
        kwargs = dict(self.getAttributeValues())
        self.parent.cropMarks = kwargs.get('useCropMarks', False)
        self.parent.pageMode = kwargs.get('pageMode')
        self.parent.pageLayout = kwargs.get('pageLayout')
        for name in self.viewerOptions:
            setattr(self.parent, name, kwargs.get(name))
        super().process()


class IDocument(interfaces.IRMLDirectiveSignature):
    occurence.containing(
        occurence.ZeroOrOne('docinit', IDocInit),
        occurence.ZeroOrOne('stylesheet', stylesheet.IStylesheet),
        occurence.ZeroOrOne('template', template.ITemplate),
        occurence.ZeroOrOne('story', template.IStory),
        occurence.ZeroOrOne('pageInfo', canvas.IPageInfo),
        occurence.ZeroOrMore('pageDrawing', canvas.IPageDrawing),
    )

    filename = attr.Text(
        title='File Name',
        description=('The default name of the output file, if no output '
                     'file was provided.'),
        required=True)

    title = attr.Text(
        title='Title',
        description=('The "Title" annotation for the PDF document.'),
        required=False)

    subject = attr.Text(
        title='Subject',
        description=('The "Subject" annotation for the PDF document.'),
        required=False)

    author = attr.Text(
        title='Author',
        description=('The "Author" annotation for the PDF document.'),
        required=False)

    creator = attr.Text(
        title='Creator',
        description=('The "Creator" annotation for the PDF document.'),
        required=False)

    debug = attr.Boolean(
        title='Debug',
        description='A flag to activate the debug output.',
        default=False,
        required=False)

    compression = attr.BooleanWithDefault(
        title='Compression',
        description=('A flag determining whether page compression should '
                     'be used.'),
        required=False)

    invariant = attr.BooleanWithDefault(
        title='Invariant',
        description=('A flag that determines whether the produced PDF '
                     'should be invariant with respect to the date and '
                     'the exact contents.'),
        required=False)


@zope.interface.implementer(interfaces.IManager,
                            interfaces.IPostProcessorManager,
                            interfaces.ICanvasManager)
class Document(directive.RMLDirective):
    signature = IDocument

    factories = {
        'docinit': DocInit,
        'stylesheet': stylesheet.Stylesheet,
        'template': template.Template,
        'story': template.Story,
        'pageInfo': canvas.PageInfo,
        'pageDrawing': canvas.PageDrawing,
    }

    def __init__(self, element, canvasClass=None):
        super().__init__(element, None)
        self.names = {}
        self.styles = {}
        self.colors = {}
        self.indexes = {}
        self.postProcessors = []
        self.filename = '<unknown>'
        self.cropMarks = False
        self.pageLayout = None
        self.pageMode = None
        self.logger = None
        self.svgs = {}
        self.attributesCache = {}
        for name in DocInit.viewerOptions:
            setattr(self, name, None)
        if not canvasClass:
            canvasClass = reportlab.pdfgen.canvas.Canvas
        self.canvasClass = canvasClass

    def _indexAdd(self, canvas, name, label):
        self.indexes[name](canvas, name, label)

    def _beforeDocument(self):
        self._initCanvas(self.doc.canv)
        self.canvas = self.doc.canv

    def _initCanvas(self, canvas):
        canvas._indexAdd = self._indexAdd
        canvas.manager = self
        if self.pageLayout:
            canvas._doc._catalog.setPageLayout(self.pageLayout)
        if self.pageMode:
            canvas._doc._catalog.setPageMode(self.pageMode)
        for name, option in DocInit.viewerOptions.items():
            if getattr(self, name) is not None:
                canvas.setViewerPreference(option, getattr(self, name))
        # Setting annotations.
        data = dict(self.getAttributeValues(
            select=('title', 'subject', 'author', 'creator')))
        canvas.setTitle(data.get('title'))
        canvas.setSubject(data.get('subject'))
        canvas.setAuthor(data.get('author'))
        canvas.setCreator(data.get('creator'))

    def process(self, outputFile=None, maxPasses=2):
        """Process document"""
        # Reset all reportlab global variables. This is very important for
        # ReportLab not to fail.
        reportlab.rl_config._reset()

        debug = self.getAttributeValues(select=('debug',), valuesOnly=True)[0]
        if not debug:
            reportlab.rl_config.shapeChecking = 0

        # Add our colors mapping to the default ones.
        colors.toColor.setExtraColorsNameSpace(self.colors)

        if outputFile is None:
            # TODO: This is relative to the input file *not* the CWD!!!
            outputFile = open(self.element.get('filename'), 'wb')

        # Create a temporary output file, so that post-processors can
        # massage the output
        self.outputFile = tempOutput = io.BytesIO()

        # Process common sub-directives
        self.processSubDirectives(select=('docinit', 'stylesheet'))

        # Handle Page Drawing Documents
        if self.element.find('pageDrawing') is not None:
            kwargs = dict(self.getAttributeValues(
                select=('compression', 'debug'),
                attrMapping={'compression': 'pageCompression',
                             'debug': 'verbosity'}
            ))
            kwargs['cropMarks'] = self.cropMarks

            self.canvas = self.canvasClass(tempOutput, **kwargs)
            self._initCanvas(self.canvas)
            self.processSubDirectives(select=('pageInfo', 'pageDrawing'))

            if hasattr(self.canvas, 'AcroForm'):
                # Makes default values appear in ReportLab >= 3.1.44
                self.canvas.AcroForm.needAppearances = 'true'

            self.canvas.save()

        # Handle Flowable-based documents.
        elif self.element.find('template') is not None:
            self.processSubDirectives(select=('template', 'story'))
            self.doc.beforeDocument = self._beforeDocument

            def callback(event, value):
                if event == 'PASS':
                    self.doc.current_pass = value

            self.doc.setProgressCallBack(callback)
            self.doc.multiBuild(
                self.flowables, maxPasses=maxPasses,
                **{'canvasmaker': self.canvasClass})

        # Process all post processors
        for name, processor in self.postProcessors:
            tempOutput.seek(0)
            tempOutput = processor.process(tempOutput)

        # Save the result into our real output file
        tempOutput.seek(0)
        outputFile.write(tempOutput.getvalue())

        # Cleanup.
        colors.toColor.setExtraColorsNameSpace({})
        reportlab.rl_config.shapeChecking = 1

    def get_name(self, name, default=None):
        if default is None:
            default = ''

        if name not in self.names:
            if self.doc._indexingFlowables and isinstance(
                self.doc._indexingFlowables[-1],
                DummyIndexingFlowable
            ):
                return default
            self.doc._indexingFlowables.append(DummyIndexingFlowable())

        return self.names.get(name, default)


class DummyIndexingFlowable(IndexingFlowable):
    """A dummy flowable to trick multiBuild into performing +1 pass."""

    def __init__(self):
        self.i = -1

    def isSatisfied(self):
        self.i += 1
        return self.i



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTLAR PURPOSE.
#
##############################################################################
"""Testing Number2Words.
"""
import unittest

from z3c.rml import num2words


class Number2WordsTests(unittest.TestCase):

    def test_num2words(self):
        for item in NUM2WORDS.strip().splitlines():
            num, expected = item.split(' ', 1)
            out = num2words.num2words(int(num))
            self.assertEqual(out, expected)

    def test_num2words_ordinal(self):
        for item in NUM2WORDS_ORDINAL.strip().splitlines():
            num, expected = item.split(' ', 1)
            out = num2words.num2words(int(num), ordinal=True)
            self.assertEqual(out, expected)


NUM2WORDS = """
0 Zero
1 One
2 Two
3 Three
4 Four
5 Five
6 Six
7 Seven
8 Eight
9 Nine
10 Ten
11 Eleven
12 Twelve
13 Thirteen
14 Fourteen
15 Fifteen
16 Sixteen
17 Seventeen
18 Eighteen
19 Nineteen
20 Twenty
21 Twenty-One
22 Twenty-Two
23 Twenty-Three
24 Twenty-Four
25 Twenty-Five
26 Twenty-Six
27 Twenty-Seven
28 Twenty-Eight
29 Twenty-Nine
30 Thirty
31 Thirty-One
40 Forty
49 Forty-Nine
50 Fifty
56 Fifty-Six
60 Sixty
66 Sixty-Six
70 Seventy
74 Seventy-Four
80 Eighty
89 Eighty-Nine
90 Ninety
91 Ninety-One
99 Ninety-Nine
100 One Hundred
101 One Hundred One
102 One Hundred Two
103 One Hundred Three
104 One Hundred Four
201 Two Hundred One
302 Three Hundred Two
403 Four Hundred Three
504 Five Hundred Four
605 Six Hundred Five
706 Seven Hundred Six
807 Eight Hundred Seven
908 Nine Hundred Eight
999 Nine Hundred Ninety-Nine
1000 One Thousand
1001 One Thousand One
1111 One Thousand One Hundred Eleven
20000 Twenty Thousand
300000 Three Hundred Thousand
4000010 Four Million Ten
4242424242 Four Billion Two Hundred Forty-Two Million Four Hundred Twenty-Four Thousand Two Hundred Forty-Two
363636363636 Three Hundred Sixty-Three Billion Six Hundred Thirty-Six Million Three Hundred Sixty-Three Thousand Six Hundred Thirty-Six
"""  # noqa: E501 line too long

NUM2WORDS_ORDINAL = """
0 Zeroth
1 First
2 Second
3 Third
4 Fourth
5 Fifth
6 Sixth
7 Seventh
8 Eighth
9 Ninth
10 Tenth
11 Eleventh
12 Twelfth
13 Thirteenth
14 Fourteenth
15 Fifteenth
16 Sixteenth
17 Seventeenth
18 Eighteenth
19 Nineteenth
20 Twentieth
21 Twenty-First
22 Twenty-Second
23 Twenty-Third
24 Twenty-Fourth
25 Twenty-Fifth
26 Twenty-Sixth
27 Twenty-Seventh
28 Twenty-Eighth
29 Twenty-Ninth
30 Thirtieth
31 Thirty-First
40 Fortieth
49 Forty-Ninth
50 Fiftieth
56 Fifty-Sixth
60 Sixtieth
66 Sixty-Sixth
70 Seventieth
74 Seventy-Fourth
80 Eightieth
89 Eighty-Ninth
90 Ninetieth
91 Ninety-First
99 Ninety-Ninth
100 One Hundredth
101 One Hundred First
102 One Hundred Second
103 One Hundred Third
104 One Hundred Fourth
201 Two Hundred First
302 Three Hundred Second
403 Four Hundred Third
504 Five Hundred Fourth
605 Six Hundred Fifth
706 Seven Hundred Sixth
807 Eight Hundred Seventh
908 Nine Hundred Eighth
999 Nine Hundred Ninety-Ninth
1000 One Thousandth
1001 One Thousand First
1111 One Thousand One Hundred Eleventh
20000 Twenty Thousandth
300000 Three Hundred Thousandth
4000010 Four Million Tenth
4242424242 Four Billion Two Hundred Forty-Two Million Four Hundred Twenty-Four Thousand Two Hundred Forty-Second
363636363636 Three Hundred Sixty-Three Billion Six Hundred Thirty-Six Million Three Hundred Sixty-Three Thousand Six Hundred Thirty-Sixth
"""  # noqa: E501 line too long



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Style Related Element Processing
"""
import copy

import reportlab.lib.enums
import reportlab.lib.styles
import reportlab.platypus

from z3c.rml import SampleStyleSheet
from z3c.rml import attr
from z3c.rml import directive
from z3c.rml import interfaces
from z3c.rml import occurence
from z3c.rml import paraparser
from z3c.rml import special


class IInitialize(interfaces.IRMLDirectiveSignature):
    """Do some RML processing initialization."""
    occurence.containing(
        occurence.ZeroOrMore('name', special.IName),
        occurence.ZeroOrMore('alias', special.IAlias),
    )


class Initialize(directive.RMLDirective):
    signature = IInitialize
    factories = {
        'name': special.Name,
        'alias': special.Alias,
    }


class ISpanStyle(interfaces.IRMLDirectiveSignature):
    """Defines a span style and gives it a name."""

    name = attr.Text(
        title='Name',
        description='The name of the style.',
        required=True)

    alias = attr.Text(
        title='Alias',
        description='An alias under which the style will also be known as.',
        required=False)

    parent = attr.Style(
        title='Parent',
        description=('The apragraph style that will be used as a base for '
                     'this one.'),
        required=False)

    fontName = attr.Text(
        title='Font Name',
        description='The name of the font for the span.',
        required=False)

    fontSize = attr.Measurement(
        title='Font Size',
        description='The font size for the text of the span.',
        required=False)

    textTransform = attr.Choice(
        title='Text Transform',
        description='Text transformation.',
        choices=interfaces.TEXT_TRANSFORM_CHOICES,
        required=False)

    underline = attr.Boolean(
        title='Underline Text',
        description='A flag, when set, causes text to be underlined.',
        required=False)

    underlineColor = attr.Color(
        title='Underline Color',
        description='The color in which the underline will appear.',
        required=False)

    underlineWidth = attr.FontSizeRelativeMeasurement(
        title='Underline Width',
        description=('The width/thickness of the underline.'),
        required=False)

    underlineOffset = attr.FontSizeRelativeMeasurement(
        title='Underline Offset',
        description=(
            'The offset of the underline with respect to the baseline.'),
        required=False)

    underlineGap = attr.FontSizeRelativeMeasurement(
        title='Underline Gap',
        description='The gap between lines for double and triple underlines.',
        required=False)

    underlineKind = attr.Choice(
        title='Underline Kind',
        description=('The kind of the underline to use.'),
        choices=interfaces.UNDERLINE_KIND_CHOICES,
        default='single',
        required=False)

    strike = attr.Boolean(
        title='Strike-through Text',
        description='A flag, when set, causes text to be struck out.',
        required=False)

    strikeColor = attr.Color(
        title='Strike Color',
        description='The color in which the strike line will appear.',
        required=False)

    strikeWidth = attr.FontSizeRelativeMeasurement(
        title='Strike Width',
        description=('The width of the strike line.'),
        required=False)

    strikeOffset = attr.FontSizeRelativeMeasurement(
        title='Strike Offset',
        description=(
            'The offset of the strike line with respect to the baseline.'),
        required=False)

    strikeGap = attr.FontSizeRelativeMeasurement(
        title='Strike Gap',
        description=(
            'The gap between lines for double and triple strike lines.'),
        required=False)

    strikeKind = attr.Choice(
        title='Strike Kind',
        description=('The kind of the strike to use.'),
        choices=interfaces.STRIKE_KIND_CHOICES,
        default='single',
        required=False)

    textColor = attr.Color(
        title='Text Color',
        description='The color in which the text will appear.',
        required=False)

    backColor = attr.Color(
        title='Background Color',
        description='The background color of the span.',
        required=False)

    linkUnderline = attr.Boolean(
        title='Underline Links',
        description=(
            'A flag, when set indicating that all links should be '
            'underlined.'),
        default=False,
        required=False)


class SpanStyle(directive.RMLDirective):
    signature = ISpanStyle

    def process(self):
        kwargs = dict(self.getAttributeValues())
        parent = kwargs.pop('parent', paraparser.SpanStyle('DefaultSpan'))
        name = kwargs.pop('name')
        style = copy.deepcopy(parent)
        style.name = name[6:] if name.startswith('style.') else name

        for name, value in kwargs.items():
            setattr(style, name, value)

        manager = attr.getManager(self)
        manager.styles[style.name] = style


class IBaseParagraphStyle(ISpanStyle):

    leading = attr.Measurement(
        title='Leading',
        description=('The height of a single paragraph line. It includes '
                     'character height.'),
        required=False)

    leftIndent = attr.Measurement(
        title='Left Indentation',
        description='General indentation on the left side.',
        required=False)

    rightIndent = attr.Measurement(
        title='Right Indentation',
        description='General indentation on the right side.',
        required=False)

    firstLineIndent = attr.Measurement(
        title='First Line Indentation',
        description='The indentation of the first line in the paragraph.',
        required=False)

    alignment = attr.Choice(
        title='Alignment',
        description='The text alignment.',
        choices=interfaces.ALIGN_CHOICES,
        required=False)

    spaceBefore = attr.Measurement(
        title='Space Before',
        description='The vertical space before the paragraph.',
        required=False)

    spaceAfter = attr.Measurement(
        title='Space After',
        description='The vertical space after the paragraph.',
        required=False)

    bulletFontName = attr.Text(
        title='Bullet Font Name',
        description='The font in which the bullet character will be rendered.',
        required=False)

    bulletFontSize = attr.Measurement(
        title='Bullet Font Size',
        description='The font size of the bullet character.',
        required=False)

    bulletIndent = attr.Measurement(
        title='Bullet Indentation',
        description='The indentation that is kept for a bullet point.',
        required=False)

    bulletColor = attr.Color(
        title='Bullet Color',
        description='The color in which the bullet will appear.',
        required=False)

    wordWrap = attr.Choice(
        title='Word Wrap Method',
        description=(
            'When set to "CJK", invoke CJK word wrapping. LTR RTL use '
            'left to right / right to left with support from pyfribi2 if '
            'available'),
        choices=interfaces.WORD_WRAP_CHOICES,
        required=False)

    borderWidth = attr.Measurement(
        title='Paragraph Border Width',
        description='The width of the paragraph border.',
        required=False)

    borderPadding = attr.Padding(
        title='Paragraph Border Padding',
        description='Padding of the paragraph.',
        required=False)

    borderColor = attr.Color(
        title='Border Color',
        description='The color in which the paragraph border will appear.',
        required=False)

    borderRadius = attr.Measurement(
        title='Paragraph Border Radius',
        description='The radius of the paragraph border.',
        required=False)

    allowWidows = attr.Boolean(
        title='Allow Widows',
        description=('Allow widows.'),
        required=False)

    allowOrphans = attr.Boolean(
        title='Allow Orphans',
        description=('Allow orphans.'),
        required=False)

    endDots = attr.Text(
        title='End Dots',
        description='Characters/Dots at the end of a paragraph.',
        required=False)

    splitLongWords = attr.Boolean(
        title='Split Long Words',
        description=('Try to split long words at the end of a line.'),
        default=True,
        required=False)

    justifyLastLine = attr.Integer(
        title='Justify Last Line',
        description=(
            'Justify last line if there are more then this number of words. '
            'Otherwise, don\'t bother.'),
        default=0,
        required=False)

    justifyBreaks = attr.Boolean(
        title='Justify Breaks',
        description=(
            'A flag, when set indicates that a line with a break should be '
            'justified as well.'),
        default=False,
        required=False)

    spaceShrinkage = attr.Float(
        title='Allowed Whitespace Shrinkage Fraction',
        description=(
            'The fraction of the original whitespace by which the '
            'whitespace is allowed to shrink to fit content on the same '
            'line.'),
        required=False)

    bulletAnchor = attr.Choice(
        title='Bullet Anchor',
        description='The place at which the bullet is anchored.',
        choices=interfaces.BULLET_ANCHOR_CHOICES,
        default='start',
        required=False)

    # Attributes not part of the official style attributes, but are accessed
    # by the paragraph renderer.

    keepWithNext = attr.Boolean(
        title='Keep with Next',
        description=('When set, this paragraph will always be in the same '
                     'frame as the following flowable.'),
        required=False)

    pageBreakBefore = attr.Boolean(
        title='Page Break Before',
        description=('Specifies whether a page break should be inserted '
                     'before the directive.'),
        required=False)

    frameBreakBefore = attr.Boolean(
        title='Frame Break Before',
        description=('Specifies whether a frame break should be inserted '
                     'before the directive.'),
        required=False)


class IParagraphStyle(IBaseParagraphStyle):
    """Defines a paragraph style and gives it a name."""

    name = attr.Text(
        title='Name',
        description='The name of the style.',
        required=True)

    alias = attr.Text(
        title='Alias',
        description='An alias under which the style will also be known as.',
        required=False)

    parent = attr.Style(
        title='Parent',
        description=('The apragraph style that will be used as a base for '
                     'this one.'),
        required=False)


class ParagraphStyle(directive.RMLDirective):
    signature = IParagraphStyle

    def process(self):
        kwargs = dict(self.getAttributeValues())
        parent = kwargs.pop(
            'parent', SampleStyleSheet['Normal'])
        name = kwargs.pop('name')
        style = copy.deepcopy(parent)
        style.name = name[6:] if name.startswith('style.') else name

        for name, value in kwargs.items():
            setattr(style, name, value)

        manager = attr.getManager(self)
        manager.styles[style.name] = style


class ITableStyleCommand(interfaces.IRMLDirectiveSignature):

    start = attr.Sequence(
        title='Start Coordinates',
        description='The start table coordinates for the style instruction',
        value_type=attr.Combination(
            value_types=(attr.Integer(),
                         attr.Choice(choices=interfaces.SPLIT_CHOICES))
        ),
        default=[0, 0],
        min_length=2,
        max_length=2,
        required=True)

    stop = attr.Sequence(
        title='End Coordinates',
        description='The end table coordinates for the style instruction',
        value_type=attr.Combination(
            value_types=(attr.Integer(),
                         attr.Choice(choices=interfaces.SPLIT_CHOICES))
        ),
        default=[-1, -1],
        min_length=2,
        max_length=2,
        required=True)


class TableStyleCommand(directive.RMLDirective):
    name = None

    def process(self):
        args = [self.name]
        args += self.getAttributeValues(valuesOnly=True)
        self.parent.style.add(*args)


class IBlockFont(ITableStyleCommand):
    """Set the font properties for the texts."""

    name = attr.Text(
        title='Font Name',
        description='The name of the font for the cell.',
        required=False)

    size = attr.Measurement(
        title='Font Size',
        description='The font size for the text of the cell.',
        required=False)

    leading = attr.Measurement(
        title='Leading',
        description=('The height of a single text line. It includes '
                     'character height.'),
        required=False)


class BlockFont(TableStyleCommand):
    signature = IBlockFont
    name = 'FONT'


class IBlockLeading(ITableStyleCommand):
    """Set the text leading."""

    length = attr.Measurement(
        title='Length',
        description=('The height of a single text line. It includes '
                     'character height.'),
        required=True)


class BlockLeading(TableStyleCommand):
    signature = IBlockLeading
    name = 'LEADING'


class IBlockTextColor(ITableStyleCommand):
    """Set the text color."""

    colorName = attr.Color(
        title='Color Name',
        description='The color in which the text will appear.',
        required=True)


class BlockTextColor(TableStyleCommand):
    signature = IBlockTextColor
    name = 'TEXTCOLOR'


class IBlockAlignment(ITableStyleCommand):
    """Set the text alignment."""

    value = attr.Choice(
        title='Text Alignment',
        description='The text alignment within the cell.',
        choices=interfaces.ALIGN_TEXT_CHOICES,
        required=True)


class BlockAlignment(TableStyleCommand):
    signature = IBlockAlignment
    name = 'ALIGNMENT'


class IBlockLeftPadding(ITableStyleCommand):
    """Set the left padding of the cells."""

    length = attr.Measurement(
        title='Length',
        description='The size of the padding.',
        required=True)


class BlockLeftPadding(TableStyleCommand):
    signature = IBlockLeftPadding
    name = 'LEFTPADDING'


class IBlockRightPadding(ITableStyleCommand):
    """Set the right padding of the cells."""

    length = attr.Measurement(
        title='Length',
        description='The size of the padding.',
        required=True)


class BlockRightPadding(TableStyleCommand):
    signature = IBlockRightPadding
    name = 'RIGHTPADDING'


class IBlockBottomPadding(ITableStyleCommand):
    """Set the bottom padding of the cells."""

    length = attr.Measurement(
        title='Length',
        description='The size of the padding.',
        required=True)


class BlockBottomPadding(TableStyleCommand):
    signature = IBlockBottomPadding
    name = 'BOTTOMPADDING'


class IBlockTopPadding(ITableStyleCommand):
    """Set the top padding of the cells."""

    length = attr.Measurement(
        title='Length',
        description='The size of the padding.',
        required=True)


class BlockTopPadding(TableStyleCommand):
    signature = IBlockTopPadding
    name = 'TOPPADDING'


class IBlockBackground(ITableStyleCommand):
    """Define the background color of the cells.

    It also supports alternating colors.
    """

    colorName = attr.Color(
        title='Color Name',
        description='The color to use as the background for every cell.',
        required=False)

    colorsByRow = attr.Sequence(
        title='Colors By Row',
        description='A list of colors to be used circularly for rows.',
        value_type=attr.Color(acceptNone=True),
        required=False)

    colorsByCol = attr.Sequence(
        title='Colors By Column',
        description='A list of colors to be used circularly for columns.',
        value_type=attr.Color(acceptNone=True),
        required=False)


class BlockBackground(TableStyleCommand):
    signature = IBlockBackground
    name = 'BACKGROUND'

    def process(self):
        args = [self.name]
        if 'colorsByRow' in self.element.keys():
            args = [BlockRowBackground.name]
        elif 'colorsByCol' in self.element.keys():
            args = [BlockColBackground.name]

        args += self.getAttributeValues(valuesOnly=True)
        self.parent.style.add(*args)


class IBlockRowBackground(ITableStyleCommand):
    """Define the background colors for rows."""

    colorNames = attr.Sequence(
        title='Colors By Row',
        description='A list of colors to be used circularly for rows.',
        value_type=attr.Color(),
        required=True)


class BlockRowBackground(TableStyleCommand):
    signature = IBlockRowBackground
    name = 'ROWBACKGROUNDS'


class IBlockColBackground(ITableStyleCommand):
    """Define the background colors for columns."""

    colorNames = attr.Sequence(
        title='Colors By Row',
        description='A list of colors to be used circularly for rows.',
        value_type=attr.Color(),
        required=True)


class BlockColBackground(TableStyleCommand):
    signature = IBlockColBackground
    name = 'COLBACKGROUNDS'


class IBlockValign(ITableStyleCommand):
    """Define the vertical alignment of the cells."""

    value = attr.Choice(
        title='Vertical Alignment',
        description='The vertical alignment of the text with the cells.',
        choices=interfaces.VALIGN_TEXT_CHOICES,
        required=True)


class BlockValign(TableStyleCommand):
    signature = IBlockValign
    name = 'VALIGN'


class IBlockSpan(ITableStyleCommand):
    """Define a span over multiple cells (rows and columns)."""


class BlockSpan(TableStyleCommand):
    signature = IBlockSpan
    name = 'SPAN'


class IBlockNosplit(ITableStyleCommand):
    """Define a nosplit over multiple cells (rows and columns)."""


class BlockNosplit(TableStyleCommand):
    signature = IBlockNosplit
    name = 'NOSPLIT'


class ILineStyle(ITableStyleCommand):
    """Define the border line style of each cell."""

    kind = attr.Choice(
        title='Kind',
        description='The kind of line actions to be taken.',
        choices=('GRID', 'BOX', 'OUTLINE', 'INNERGRID',
                 'LINEBELOW', 'LINEABOVE', 'LINEBEFORE', 'LINEAFTER'),
        required=True)

    thickness = attr.Measurement(
        title='Thickness',
        description='Line Thickness',
        default=1,
        required=True)

    colorName = attr.Color(
        title='Color',
        description='The color of the border line.',
        default=None,
        required=True)

    cap = attr.Choice(
        title='Cap',
        description='The cap at the end of a border line.',
        choices=interfaces.CAP_CHOICES,
        default=1,
        required=True)

    dash = attr.Sequence(
        title='Dash-Pattern',
        description='The dash-pattern of a line.',
        value_type=attr.Measurement(),
        default=None,
        required=False)

    join = attr.Choice(
        title='Join',
        description='The way lines are joined together.',
        choices=interfaces.JOIN_CHOICES,
        default=1,
        required=False)

    count = attr.Integer(
        title='Count',
        description=('Describes whether the line is a single (1) or '
                     'double (2) line.'),
        default=1,
        required=False)


class LineStyle(TableStyleCommand):
    signature = ILineStyle

    def process(self):
        name = self.getAttributeValues(select=('kind',), valuesOnly=True)[0]
        args = [name]
        args += self.getAttributeValues(ignore=('kind',), valuesOnly=True,
                                        includeMissing=True)
        args = [val if val is not attr.MISSING else None for val in args]
        self.parent.style.add(*args)


class IBlockTableStyle(interfaces.IRMLDirectiveSignature):
    """A style defining the look of a table."""
    occurence.containing(
        occurence.ZeroOrMore('blockFont', IBlockFont),
        occurence.ZeroOrMore('blockLeading', IBlockLeading),
        occurence.ZeroOrMore('blockTextColor', IBlockTextColor),
        occurence.ZeroOrMore('blockAlignment', IBlockAlignment),
        occurence.ZeroOrMore('blockLeftPadding', IBlockLeftPadding),
        occurence.ZeroOrMore('blockRightPadding', IBlockRightPadding),
        occurence.ZeroOrMore('blockBottomPadding', IBlockBottomPadding),
        occurence.ZeroOrMore('blockTopPadding', IBlockTopPadding),
        occurence.ZeroOrMore('blockBackground', IBlockBackground),
        occurence.ZeroOrMore('blockRowBackground', IBlockRowBackground),
        occurence.ZeroOrMore('blockColBackground', IBlockColBackground),
        occurence.ZeroOrMore('blockValign', IBlockValign),
        occurence.ZeroOrMore('blockSpan', IBlockSpan),
        occurence.ZeroOrMore('blockNosplit', IBlockNosplit),
        occurence.ZeroOrMore('lineStyle', ILineStyle)
    )

    id = attr.Text(
        title='Id',
        description='The name/id of the style.',
        required=True)

    keepWithNext = attr.Boolean(
        title='Keep with Next',
        description=('When set, this paragraph will always be in the same '
                     'frame as the following flowable.'),
        required=False)


class BlockTableStyle(directive.RMLDirective):
    signature = IBlockTableStyle

    factories = {
        'blockFont': BlockFont,
        'blockLeading': BlockLeading,
        'blockTextColor': BlockTextColor,
        'blockAlignment': BlockAlignment,
        'blockLeftPadding': BlockLeftPadding,
        'blockRightPadding': BlockRightPadding,
        'blockBottomPadding': BlockBottomPadding,
        'blockTopPadding': BlockTopPadding,
        'blockBackground': BlockBackground,
        'blockRowBackground': BlockRowBackground,
        'blockColBackground': BlockColBackground,
        'blockValign': BlockValign,
        'blockSpan': BlockSpan,
        'blockNosplit': BlockNosplit,
        'lineStyle': LineStyle,
    }

    def process(self):
        kw = dict(self.getAttributeValues())
        id = kw.pop('id')
        # Create Style
        self.style = reportlab.platypus.tables.TableStyle()
        for name, value in kw.items():
            setattr(self.style, name, value)
        # Fill style
        self.processSubDirectives()
        # Add style to the manager
        manager = attr.getManager(self)
        manager.styles[id] = self.style


class IMinimalListStyle(interfaces.IRMLDirectiveSignature):

    leftIndent = attr.Measurement(
        title='Left Indentation',
        description='General indentation on the left side.',
        required=False)

    rightIndent = attr.Measurement(
        title='Right Indentation',
        description='General indentation on the right side.',
        required=False)

    bulletColor = attr.Color(
        title='Bullet Color',
        description='The color in which the bullet will appear.',
        required=False)

    bulletFontName = attr.Text(
        title='Bullet Font Name',
        description='The font in which the bullet character will be rendered.',
        required=False)

    bulletFontSize = attr.Measurement(
        title='Bullet Font Size',
        description='The font size of the bullet character.',
        required=False)

    bulletOffsetY = attr.Measurement(
        title='Bullet Y-Offset',
        description='The vertical offset of the bullet.',
        required=False)

    bulletDedent = attr.StringOrInt(
        title='Bullet Dedent',
        description='Either pixels of dedent or auto (default).',
        required=False)

    bulletDir = attr.Choice(
        title='Bullet Layout Direction',
        description='The layout direction of the bullet.',
        choices=('ltr', 'rtl'),
        required=False)

    bulletFormat = attr.Text(
        title='Bullet Format',
        description='A formatting expression for the bullet text.',
        required=False)

    bulletType = attr.Choice(
        title='Bullet Type',
        description='The type of number to display.',
        choices=interfaces.ORDERED_LIST_TYPES +
        interfaces.UNORDERED_BULLET_VALUES,
        doLower=False,
        required=False)


class IBaseListStyle(IMinimalListStyle):

    start = attr.Combination(
        title='Start Value',
        description='The counter start value.',
        value_types=(
            # Numeric start value.
            attr.Integer(),
            # Bullet types.
            attr.Choice(choices=interfaces.UNORDERED_BULLET_VALUES),
            # Arbitrary text.
            attr.Text(),
        ),
        required=False)


class IListStyle(IBaseListStyle):
    """Defines a list style and gives it a name."""

    name = attr.Text(
        title='Name',
        description='The name of the style.',
        required=True)

    parent = attr.Style(
        title='Parent',
        description=('The list style that will be used as a base for '
                     'this one.'),
        required=False)


class ListStyle(directive.RMLDirective):
    signature = IListStyle

    def process(self):
        kwargs = dict(self.getAttributeValues())
        parent = kwargs.pop(
            'parent', reportlab.lib.styles.ListStyle(name='List'))
        name = kwargs.pop('name')
        style = copy.deepcopy(parent)
        style.name = name[6:] if name.startswith('style.') else name

        for name, value in kwargs.items():
            setattr(style, name, value)

        manager = attr.getManager(self)
        manager.styles[style.name] = style


class IStylesheet(interfaces.IRMLDirectiveSignature):
    """A styleheet defines the styles that can be used in the document."""
    occurence.containing(
        occurence.ZeroOrOne('initialize', IInitialize),
        occurence.ZeroOrMore('spanStyle', ISpanStyle),
        occurence.ZeroOrMore('paraStyle', IParagraphStyle),
        occurence.ZeroOrMore('blockTableStyle', IBlockTableStyle),
        occurence.ZeroOrMore('listStyle', IListStyle),
        # TODO:
        # occurence.ZeroOrMore('boxStyle', IBoxStyle),
    )


class Stylesheet(directive.RMLDirective):
    signature = IStylesheet

    factories = {
        'initialize': Initialize,
        'spanStyle': SpanStyle,
        'paraStyle': ParagraphStyle,
        'blockTableStyle': BlockTableStyle,
        'listStyle': ListStyle,
    }



###############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
###############################################################################
"""Tests for the Book Documentation Module"""

import doctest


bad_rml = '''\
<document filename="bad.pdf">
  <bad />
</document>
'''


def test_abort_on_invalid_tag():
    """

    We can set a flag that aborts execution when a bad tag/directive is
    encountered:

      >>> from z3c.rml import directive
      >>> directive.ABORT_ON_INVALID_DIRECTIVE = True

    Let's now render a template:

      >>> from z3c.rml import rml2pdf
      >>> rml2pdf.parseString(bad_rml, True, 'bad.rml')
      Traceback (most recent call last):
      ...
      ValueError: Directive 'bad' could not be processed and was ignored.
      (file bad.rml, line 2)

    Cleanup:

      >>> directive.ABORT_ON_INVALID_DIRECTIVE = False

    """


def test_suite():
    return doctest.DocTestSuite(
        optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS)



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""RML to PDF Converter
"""
import os
import subprocess
import sys


_fileOpen = None


def excecuteSubProcess(xmlInputName, outputFileName, testing=None):
    # set the sys path given from the parent process
    sysPath = os.environ['Z3CRMLSYSPATH']
    sys.path[:] = sysPath.split(';')

    # now it come the ugly thing, but we need to hook our test changes into
    # our subprocess.
    if testing is not None:

        # set some globals
        import z3c.rml.attr
        import z3c.rml.directive
        global _fileOpen
        _fileOpen = z3c.rml.attr.File.open

        def testOpen(img, filename):
            # cleanup win paths like:
            # ....\\input\\file:///D:\\trunk\\...
            if sys.platform[:3].lower() == "win":
                if filename.startswith('file:///'):
                    filename = filename[len('file:///'):]
            path = os.path.join(os.path.dirname(xmlInputName), filename)
            return open(path, 'rb')
        # override some testing stuff for our tests
        z3c.rml.attr.File.open = testOpen
        import z3c.rml.tests.module
        sys.modules['module'] = z3c.rml.tests.module
        sys.modules['mymodule'] = z3c.rml.tests.module

    # import rml and process the pdf
    from z3c.rml import rml2pdf
    rml2pdf.go(xmlInputName, outputFileName)

    if testing is not None:
        # reset some globals
        z3c.rml.attr.File.open = _fileOpen
        del sys.modules['module']
        del sys.modules['mymodule']


def goSubProcess(xmlInputName, outputFileName, testing=False):
    """Processes PDF rendering in a sub process.

    This method is much slower then the ``go`` method defined in rml2pdf.py
    because each PDF generation is done in a sub process. But this will make
    sure, that we do not run into problems. Note, the ReportLab lib is not
    threadsafe.

    Use this method from python and it will dispatch the pdf generation
    to a subprocess.

    Note: this method does not take care on how much process will started.
    Probably it's a good idea to use a queue or a global utility which only
    start a predefined amount of sub processes.
    """
    # get the sys path used for this python process
    env = os.environ
    sysPath = ';'.join(sys.path)
    # set the sys path as env var for the new sub process
    env['Z3CRMLSYSPATH'] = sysPath
    py = sys.executable

    # setup the cmd
    program = [
        py,
        __file__,
        'excecuteSubProcess',
        xmlInputName,
        outputFileName]
    if testing is True:
        program.append('testing=1')
    program = " ".join(program)

    # run the subprocess in the rml input file folder, this will make it easy
    # to include images. If this doesn't fit, feel free to add a additional
    # home argument, and let this be the default, ri
    os.chdir(os.path.dirname(xmlInputName))

    # start processing in a sub process, raise exception or return None
    try:
        p = subprocess.Popen(program, executable=py, env=env,
                             stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
    except Exception as e:
        raise Exception("Subprocess error: %s" % e)

    # Do we need to improve the implementation and kill the subprocess which
    # will fail? ri
    stdout, stderr = p.communicate()
    error = stderr
    if error:
        raise Exception("Subprocess error: %s" % error)


if __name__ == '__main__':
    if len(sys.argv) == 5:
        # testing support
        canvas = excecuteSubProcess(sys.argv[2], sys.argv[3], sys.argv[4])
    else:
        canvas = excecuteSubProcess(sys.argv[2], sys.argv[3])



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTLAR PURPOSE.
#
##############################################################################
"""Test RML to PDF converter.
"""

import io
import unittest
from unittest import mock

from z3c.rml import rml2pdf


class RML2PDFTest(unittest.TestCase):

    @mock.patch("z3c.rml.rml2pdf.go")
    def test_main_minimal(self, go):
        rml2pdf.main(["input.rml", "output.pdf"])
        go.assert_called_with("input.rml", "output.pdf")

    @mock.patch("z3c.rml.rml2pdf.go")
    def test_main_all_args(self, go):
        rml2pdf.main(["input.rml", "output.pdf", "./out/pdf/", "./out/dtd/"])
        go.assert_called_with(
            "input.rml", "output.pdf", "./out/pdf/", "./out/dtd/"
        )

    def test_go(self):
        rml = """
            <!DOCTYPE document SYSTEM "rml_1_0.dtd">
            <document filename="test.pdf">
            </document>
        """
        rml2pdf.go(io.StringIO(rml))

    def test_parseString(self):
        rml = """
          <?xml version="1.0" encoding="UTF-8" ?>
          <!DOCTYPE document SYSTEM "rml_1_0.dtd">
          <document filename="test.pdf">
            <template>
              <pageTemplate id="main">
                <frame id="first" x1="1in" y1="1in"
                       width="7in" height="9in"/>
              </pageTemplate>
            </template>
            <story><para>Hello</para></story>
          </document>
        """.strip()
        stream = rml2pdf.parseString(rml)
        self.assertEqual(stream.read()[:8], b"%PDF-1.4")



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Page Drawing Related Element Processing
"""
import reportlab.pdfgen.canvas  # noqa: F401 imported but unused

from z3c.rml import attr
from z3c.rml import chart
from z3c.rml import directive
from z3c.rml import flowable
from z3c.rml import form
from z3c.rml import interfaces
from z3c.rml import occurence
from z3c.rml import page
from z3c.rml import special
from z3c.rml import stylesheet  # noqa: F401 imported but unused


class IShape(interfaces.IRMLDirectiveSignature):
    """A shape to be drawn on the canvas."""

    x = attr.Measurement(
        title='X-Coordinate',
        description=('The X-coordinate of the lower-left position of the '
                     'shape.'),
        required=True)

    y = attr.Measurement(
        title='Y-Coordinate',
        description=('The Y-coordinate of the lower-left position of the '
                     'shape.'),
        required=True)

    fill = attr.Boolean(
        title='Fill',
        description='A flag to specify whether the shape should be filled.',
        required=False)

    stroke = attr.Boolean(
        title='Stroke',
        description=("A flag to specify whether the shape's outline should "
                     "be drawn."),
        required=False)


class CanvasRMLDirective(directive.RMLDirective):
    callable = None
    attrMapping = None

    def process(self):
        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        getattr(canvas, self.callable)(**kwargs)


class ISaveState(interfaces.IRMLDirectiveSignature):
    """Saves the current canvas state."""


class SaveState(CanvasRMLDirective):
    signature = ISaveState
    callable = 'saveState'


class IRestoreState(interfaces.IRMLDirectiveSignature):
    """Saves the current canvas state."""


class RestoreState(CanvasRMLDirective):
    signature = IRestoreState
    callable = 'restoreState'


class IDrawString(interfaces.IRMLDirectiveSignature):
    """Draws a simple string (left aligned) onto the canvas at the specified
    location."""

    x = attr.Measurement(
        title='X-Coordinate',
        description=('The X-coordinate of the lower-left position of the '
                     'string.'),
        required=True)

    y = attr.Measurement(
        title='Y-Coordinate',
        description=('The Y-coordinate of the lower-left position of the '
                     'string.'),
        required=True)

    text = attr.RawXMLContent(
        title='Text',
        description=('The string/text that is put onto the canvas.'),
        required=True)


class DrawString(CanvasRMLDirective, special.TextFlowables):
    signature = IDrawString
    callable = 'drawString'

    def process(self):
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        kwargs['text'] = self._getText(self.element, canvas).strip()
        getattr(canvas, self.callable)(**kwargs)


class IDrawRightString(IDrawString):
    """Draws a simple string (right aligned) onto the canvas at the specified
    location."""


class DrawRightString(DrawString):
    signature = IDrawRightString
    callable = 'drawRightString'


class IDrawCenteredString(IDrawString):
    """Draws a simple string (centered aligned) onto the canvas at the
    specified location."""


class DrawCenteredString(DrawString):
    signature = IDrawCenteredString
    callable = 'drawCentredString'


class IDrawAlignedString(IDrawString):
    """Draws a simple string (aligned to the pivot character) onto the canvas
    at the specified location."""

    pivotChar = attr.Text(
        title='Text',
        description=('The string/text that is put onto the canvas.'),
        min_length=1,
        max_length=1,
        default='.',
        required=True)


class DrawAlignedString(DrawString):
    signature = IDrawAlignedString
    callable = 'drawAlignedString'


class IEllipse(IShape):
    """Draws an ellipse on the canvas."""

    width = attr.Measurement(
        title='Width',
        description='The width of the ellipse.',
        required=True)

    height = attr.Measurement(
        title='Height',
        description='The height of the ellipse.',
        required=True)


class Ellipse(CanvasRMLDirective):
    signature = IEllipse
    callable = 'ellipse'
    attrMapping = {'x': 'x1', 'y': 'y1'}

    def process(self):
        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        # Convert width and height to end locations
        kwargs['x2'] = kwargs['x1'] + kwargs['width']
        del kwargs['width']
        kwargs['y2'] = kwargs['y1'] + kwargs['height']
        del kwargs['height']
        getattr(canvas, self.callable)(**kwargs)


class ICircle(IShape):
    """Draws a circle on the canvas."""

    radius = attr.Measurement(
        title='Radius',
        description='The radius of the circle.',
        required=True)


class Circle(CanvasRMLDirective):
    signature = ICircle
    callable = 'circle'
    attrMapping = {'x': 'x_cen', 'y': 'y_cen', 'radius': 'r'}


class IRectangle(IShape):
    """Draws an ellipse on the canvas."""

    width = attr.Measurement(
        title='Width',
        description='The width of the rectangle.',
        required=True)

    height = attr.Measurement(
        title='Height',
        description='The height of the rectangle.',
        required=True)

    round = attr.Measurement(
        title='Corner Radius',
        description='The radius of the rounded corners.',
        required=False)

    href = attr.Text(
        title='Link URL',
        description='When specified, the rectangle becomes a link to that'
                    ' URL.',
        required=False)

    destination = attr.Text(
        title='Link Destination',
        description=('When specified, the rectangle becomes a link to that '
                     'destination.'),
        required=False)


class Rectangle(CanvasRMLDirective):
    signature = IRectangle
    callable = 'rect'
    attrMapping = {'round': 'radius'}

    def process(self):
        if 'round' in self.element.keys():
            self.callable = 'roundRect'
        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        # Create a link
        url = kwargs.pop('href', None)
        if url:
            canvas.linkURL(
                url,
                (kwargs['x'],
                 kwargs['y'],
                 kwargs['x'] + kwargs['width'],
                 kwargs['y'] + kwargs['height']))
        dest = kwargs.pop('destination', None)
        if dest:
            canvas.linkRect(
                '',
                dest,
                (kwargs['x'],
                 kwargs['y'],
                 kwargs['x'] + kwargs['width'],
                 kwargs['y'] + kwargs['height']))

        # Render the rectangle
        getattr(canvas, self.callable)(**kwargs)


class IGrid(interfaces.IRMLDirectiveSignature):
    """A shape to be drawn on the canvas."""

    xs = attr.Sequence(
        title='X-Coordinates',
        description=('A sequence x-coordinates that represent the vertical '
                     'line positions.'),
        value_type=attr.Measurement(),
        required=True)

    ys = attr.Sequence(
        title='Y-Coordinates',
        description=('A sequence y-coordinates that represent the horizontal '
                     'line positions.'),
        value_type=attr.Measurement(),
        required=True)


class Grid(CanvasRMLDirective):
    signature = IGrid
    callable = 'grid'
    attrMapping = {'xs': 'xlist', 'ys': 'ylist'}


class ILines(interfaces.IRMLDirectiveSignature):
    """A path of connected lines drawn on the canvas."""

    linelist = attr.TextNodeGrid(
        title='Line List',
        description=('A list of lines coordinates to draw.'),
        value_type=attr.Measurement(),
        columns=4,
        required=True)


class Lines(CanvasRMLDirective):
    signature = ILines
    callable = 'lines'


class ICurves(interfaces.IRMLDirectiveSignature):
    """A path of connected bezier curves drawn on the canvas."""

    curvelist = attr.TextNodeGrid(
        title='Curve List',
        description=('A list of curve coordinates to draw.'),
        value_type=attr.Measurement(),
        columns=8,
        required=True)


class Curves(CanvasRMLDirective):
    signature = ICurves
    callable = 'bezier'

    def process(self):
        argset = self.getAttributeValues(valuesOnly=True)[0]
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        for args in argset:
            getattr(canvas, self.callable)(*args)


class IImage(interfaces.IRMLDirectiveSignature):
    """Draws an external image on the canvas."""

    file = attr.Image(
        title='File',
        description=('Reference to the external file of the iamge.'),
        required=True)

    x = attr.Measurement(
        title='X-Coordinate',
        description=('The X-coordinate of the lower-left position of the '
                     'shape.'),
        required=True)

    y = attr.Measurement(
        title='Y-Coordinate',
        description=('The Y-coordinate of the lower-left position of the '
                     'shape.'),
        required=True)

    width = attr.Measurement(
        title='Width',
        description='The width of the image.',
        required=False)

    height = attr.Measurement(
        title='Height',
        description='The height of the image.',
        required=False)

    showBoundary = attr.Boolean(
        title='Show Boundary',
        description=('A flag determining whether a border should be drawn '
                     'around the image.'),
        default=False,
        required=False)

    preserveAspectRatio = attr.Boolean(
        title='Preserve Aspect Ratio',
        description=("A flag determining whether the image's aspect ration "
                     "should be conserved under any circumstances."),
        default=False,
        required=False)

    mask = attr.Color(
        title='Mask',
        description='The color mask used to render the image, or "auto" to use'
                    ' the alpha channel if available.',
        default='auto',
        required=False,
        acceptAuto=True)

    anchor = attr.Choice(
        title='Text Anchor',
        description='The position in the text to which the coordinates refer.',
        choices='''nw   n   ne
                w    c    e
                sw   s   se'''.split(),
        required=False)


class Image(CanvasRMLDirective):
    signature = IImage
    callable = 'drawImage'
    attrMapping = {'file': 'image'}

    def process(self):
        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        show = kwargs.pop('showBoundary')

        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        getattr(canvas, self.callable)(**kwargs)

        if show:
            width = kwargs.get('width', kwargs['image'].getSize()[0])
            height = kwargs.get('height', kwargs['image'].getSize()[1])
            canvas.rect(kwargs['x'], kwargs['y'], width, height)


class IPlace(interfaces.IRMLDirectiveSignature):
    """Draws a set of flowables on the canvas within a given region."""

    x = attr.Measurement(
        title='X-Coordinate',
        description=('The X-coordinate of the lower-left position of the '
                     'place.'),
        required=True)

    y = attr.Measurement(
        title='Y-Coordinate',
        description=('The Y-coordinate of the lower-left position of the '
                     'place.'),
        required=True)

    width = attr.Measurement(
        title='Width',
        description='The width of the place.',
        required=False)

    height = attr.Measurement(
        title='Height',
        description='The height of the place.',
        required=False)


class Place(CanvasRMLDirective):
    signature = IPlace

    def process(self):
        x, y, width, height = self.getAttributeValues(
            select=('x', 'y', 'width', 'height'), valuesOnly=True)
        y += height

        flows = flowable.Flow(self.element, self.parent)
        flows.process()

        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        for flow in flows.flow:
            flow.canv = canvas
            flowWidth, flowHeight = flow.wrap(width, height)
            if flowWidth <= width and flowHeight <= height:
                y -= flowHeight
                flow.drawOn(canvas, x, y)
                height -= flowHeight
            else:
                raise ValueError("Not enough space")


class IParam(interfaces.IRMLDirectiveSignature):
    """Sets one paramter for the text annotation."""

    name = attr.Text(
        title='Name',
        description='The name of the paramter.',
        required=True)

    value = attr.TextNode(
        title='Value',
        description=('The parameter value.'),
        required=True)


class Param(directive.RMLDirective):
    signature = IParam

    def process(self):
        args = dict(self.getAttributeValues())
        self.parent.params[args['name']] = args['value']


class ITextAnnotation(interfaces.IRMLDirectiveSignature):
    """Writes a low-level text annotation into the PDF."""
    occurence.containing(
        occurence.ZeroOrMore('param', IParam))

    contents = attr.FirstLevelTextNode(
        title='Contents',
        description='The PDF commands that are inserted as annotation.',
        required=True)


class TextAnnotation(CanvasRMLDirective):
    signature = ITextAnnotation
    factories = {'param': Param}

    paramTypes = {'escape': attr.Integer()}

    def process(self):
        contents = self.getAttributeValues(valuesOnly=True)[0]
        self.params = {}
        self.processSubDirectives()
        for name, type in self.paramTypes.items():
            if name in self.params:
                bound = type.bind(self)
                self.params[name] = bound.fromUnicode(self.params[name])
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        canvas.textAnnotation(contents, **self.params)


class IMoveTo(interfaces.IRMLDirectiveSignature):
    """Move the path cursor to the specified location."""

    position = attr.TextNodeSequence(
        title='Position',
        description='Position to which the path pointer is moved to.',
        value_type=attr.Measurement(),
        min_length=2,
        max_length=2,
        required=True)


class MoveTo(directive.RMLDirective):
    signature = IMoveTo

    def process(self):
        args = self.getAttributeValues(valuesOnly=True)
        self.parent.path.moveTo(*args[0])


class ICurveTo(interfaces.IRMLDirectiveSignature):
    """Create a bezier curve from the current location to the specified one."""

    curvelist = attr.TextNodeGrid(
        title='Curve Specification',
        description='Describes the end position and the curve properties.',
        value_type=attr.Measurement(),
        columns=6,
        required=True)


class CurveTo(directive.RMLDirective):
    signature = ICurveTo

    def process(self):
        argset = self.getAttributeValues(valuesOnly=True)[0]
        for args in argset:
            self.parent.path.curveTo(*args)


class ICurvesTo(ICurveTo):
    pass


directive.DeprecatedDirective(
    ICurvesTo,
    'Available for ReportLab RML compatibility. Please use the "curveto" '
    'directive instead.')


class IPath(IShape):
    """Create a line path."""
    occurence.containing(
        occurence.ZeroOrMore('moveto', IMoveTo),
        occurence.ZeroOrMore('curveto', ICurveTo),
        occurence.ZeroOrMore('curvesto', ICurvesTo),
    )

    points = attr.TextNodeGrid(
        title='Points',
        description=('A list of coordinate points that define th path.'),
        value_type=attr.Measurement(),
        columns=2,
        required=True)

    close = attr.Boolean(
        title='Close Path',
        description=("A flag specifying whether the path should be closed."),
        default=False,
        required=False)

    clip = attr.Boolean(
        title='Clip Path',
        description=("A flag specifying whether the path should clip "
                     "overlapping elements."),
        default=False,
        required=False)


class Path(CanvasRMLDirective):
    signature = IPath
    factories = {
        'moveto': MoveTo,
        'curveto': CurveTo,
        'curvesto': CurveTo
    }

    def processPoints(self, text):
        if text.strip() == '':
            return
        bound = self.signature['points'].bind(self)
        for coords in bound.fromUnicode(text):
            self.path.lineTo(*coords)

    def process(self):
        kwargs = dict(self.getAttributeValues(ignore=('points',)))

        # Start the path and set the cursor to the start location.
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        self.path = canvas.beginPath()
        self.path.moveTo(kwargs.pop('x'), kwargs.pop('y'))

        # Process the text before the first sub-directive.
        if self.element.text is not None:
            self.processPoints(self.element.text)
        # Handle each sub-directive.
        for directive_ in self.element.getchildren():
            if directive_.tag in self.factories:
                self.factories[directive_.tag](directive_, self).process()
            # If there is more text after sub-directive, process it.
            if directive_.tail is not None:
                self.processPoints(directive_.tail)

        if kwargs.pop('close', False):
            self.path.close()

        if kwargs.pop('clip', False):
            canvas.clipPath(self.path, **kwargs)
        else:
            canvas.drawPath(self.path, **kwargs)


class IFill(interfaces.IRMLDirectiveSignature):
    """Set the fill color."""

    color = attr.Color(
        title='Color',
        description=('The color value to be set.'),
        required=True)


class Fill(CanvasRMLDirective):
    signature = IFill
    callable = 'setFillColor'
    attrMapping = {'color': 'aColor'}


class IStroke(interfaces.IRMLDirectiveSignature):
    """Set the stroke/line color."""

    color = attr.Color(
        title='Color',
        description=('The color value to be set.'),
        required=True)


class Stroke(CanvasRMLDirective):
    signature = IStroke
    callable = 'setStrokeColor'
    attrMapping = {'color': 'aColor'}


class ISetFont(interfaces.IRMLDirectiveSignature):
    """Set the font name and/or size."""

    name = attr.Text(
        title='Font Name',
        description=('The name of the font as it was registered.'),
        required=True)

    size = attr.Measurement(
        title='Size',
        description=('The font size.'),
        required=True)

    leading = attr.Measurement(
        title='Leading',
        description=('The font leading.'),
        required=False)


class SetFont(CanvasRMLDirective):
    signature = ISetFont
    callable = 'setFont'
    attrMapping = {'name': 'psfontname'}


class ISetFontSize(interfaces.IRMLDirectiveSignature):
    """Set the font size."""

    size = attr.Measurement(
        title='Size',
        description=('The font size.'),
        required=True)

    leading = attr.Measurement(
        title='Leading',
        description=('The font leading.'),
        required=False)


class SetFontSize(CanvasRMLDirective):
    signature = ISetFontSize
    callable = 'setFontSize'


class IScale(interfaces.IRMLDirectiveSignature):
    """Scale the drawing using x and y scaling factors."""

    sx = attr.Float(
        title='X-Scaling-Factor',
        description=('The scaling factor applied on x-coordinates.'),
        default=1,
        required=False)

    sy = attr.Float(
        title='Y-Scaling-Factor',
        description=('The scaling factor applied on y-coordinates.'),
        default=1,
        required=False)


class Scale(CanvasRMLDirective):
    signature = IScale
    callable = 'scale'
    attrMapping = {'sx': 'x', 'sy': 'y'}


class ITranslate(interfaces.IRMLDirectiveSignature):
    """Translate the drawing coordinates by the specified x and y offset."""

    dx = attr.Measurement(
        title='X-Offset',
        description=('The amount to move the drawing to the right.'),
        required=True)

    dy = attr.Measurement(
        title='Y-Offset',
        description=('The amount to move the drawing upward.'),
        required=True)


class Translate(CanvasRMLDirective):
    signature = ITranslate
    callable = 'translate'


class IRotate(interfaces.IRMLDirectiveSignature):
    """Rotate the drawing counterclockwise."""

    degrees = attr.Measurement(
        title='Angle',
        description=('The angle in degrees.'),
        required=True)


class Rotate(CanvasRMLDirective):
    signature = IRotate
    callable = 'rotate'
    attrMapping = {'degrees': 'theta'}


class ISkew(interfaces.IRMLDirectiveSignature):
    """Skew the drawing."""

    alpha = attr.Measurement(
        title='Alpha',
        description=('The amount to skew the drawing in the horizontal.'),
        required=True)

    beta = attr.Measurement(
        title='Beta',
        description=('The amount to skew the drawing in the vertical.'),
        required=True)


class Skew(CanvasRMLDirective):
    signature = ISkew
    callable = 'skew'


class ITransform(interfaces.IRMLDirectiveSignature):
    """A full 2-D matrix transformation"""

    matrix = attr.TextNodeSequence(
        title='Matrix',
        description='The transformation matrix.',
        value_type=attr.Float(),
        min_length=6,
        max_length=6,
        required=True)


class Transform(CanvasRMLDirective):
    signature = ITransform

    def process(self):
        args = self.getAttributeValues(valuesOnly=True)
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        canvas.transform(*args[0])


class ILineMode(interfaces.IRMLDirectiveSignature):
    """Set the line mode for the following graphics elements."""

    width = attr.Measurement(
        title='Width',
        description=('The line width.'),
        required=False)

    dash = attr.Sequence(
        title='Dash-Pattern',
        description=('The dash-pattern of a line.'),
        value_type=attr.Measurement(),
        required=False)

    miterLimit = attr.Measurement(
        title='Miter Limit',
        description=('The ???.'),
        required=False)

    join = attr.Choice(
        title='Join',
        description='The way lines are joined together.',
        choices=interfaces.JOIN_CHOICES,
        required=False)

    cap = attr.Choice(
        title='Cap',
        description='The cap is the desciption of how the line-endings look.',
        choices=interfaces.CAP_CHOICES,
        required=False)


class LineMode(CanvasRMLDirective):
    signature = ILineMode

    def process(self):
        kw = dict(self.getAttributeValues())
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        if 'width' in kw:
            canvas.setLineWidth(kw['width'])
        if 'join' in kw:
            canvas.setLineJoin(kw['join'])
        if 'cap' in kw:
            canvas.setLineCap(kw['cap'])
        if 'miterLimit' in kw:
            canvas.setMiterLimit(kw['miterLimit'])
        if 'dash' in kw:
            canvas.setDash(kw['dash'])


class IBookmark(interfaces.IRMLDirectiveSignature):
    """
    This creates a bookmark to the current page which can be referred to with
    the given key elsewhere. (Used inside a page drawing.)
    """

    name = attr.Text(
        title='Name',
        description='The name of the bookmark.',
        required=True)

    fit = attr.Choice(
        title='Fit',
        description='The Fit Type.',
        choices=('XYZ', 'Fit', 'FitH', 'FitV', 'FitR'),
        required=False)

    zoom = attr.Float(
        title='Zoom',
        description='The zoom level when clicking on the bookmark.',
        required=False)

    x = attr.Measurement(
        title='X-Position',
        description='The x-position.',
        required=False)

    y = attr.Measurement(
        title='Y-Position',
        description='The y-position.',
        required=False)


class Bookmark(CanvasRMLDirective):
    signature = IBookmark

    def process(self):
        args = dict(self.getAttributeValues())
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        args['left'], args['top'] = canvas.absolutePosition(
            args['x'], args['y'])
        canvas.bookmarkPage(**args)


class IPlugInGraphic(interfaces.IRMLDirectiveSignature):
    """Inserts a custom graphic developed in Python."""

    module = attr.Text(
        title='Module',
        description='The Python module in which the flowable is located.',
        required=True)

    function = attr.Text(
        title='Function',
        description=('The name of the factory function within the module '
                     'that returns the custom flowable.'),
        required=True)

    params = attr.TextNode(
        title='Parameters',
        description=('A list of parameters encoded as a long string.'),
        required=False)


class PlugInGraphic(CanvasRMLDirective):
    signature = IPlugInGraphic

    def process(self):
        modulePath, functionName, params = self.getAttributeValues(
            valuesOnly=True)
        module = __import__(modulePath, {}, {}, [modulePath])
        function = getattr(module, functionName)
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        function(canvas, params)


class IDrawing(interfaces.IRMLDirectiveSignature):
    """A container directive for all directives that draw directly on the
    cnavas."""
    occurence.containing(
        # State Manipulation
        occurence.ZeroOrMore('saveState', ISaveState),
        occurence.ZeroOrMore('restoreState', IRestoreState),
        # String Drawing
        occurence.ZeroOrMore('drawString', IDrawString),
        occurence.ZeroOrMore('drawRightString', IDrawRightString),
        occurence.ZeroOrMore('drawCenteredString', IDrawCenteredString),
        occurence.ZeroOrMore('drawCentredString', IDrawCenteredString),
        occurence.ZeroOrMore('drawAlignedString', IDrawAlignedString),
        # Drawing Operations
        occurence.ZeroOrMore('ellipse', IEllipse),
        occurence.ZeroOrMore('circle', ICircle),
        occurence.ZeroOrMore('rect', IRectangle),
        occurence.ZeroOrMore('grid', IGrid),
        occurence.ZeroOrMore('lines', ILines),
        occurence.ZeroOrMore('curves', ICurves),
        occurence.ZeroOrMore('image', IImage),
        occurence.ZeroOrMore('place', IPlace),
        occurence.ZeroOrMore('textAnnotation', ITextAnnotation),
        occurence.ZeroOrMore('path', IPath),
        # State Change Operations
        occurence.ZeroOrMore('fill', IFill),
        occurence.ZeroOrMore('stroke', IStroke),
        occurence.ZeroOrMore('setFont', ISetFont),
        occurence.ZeroOrMore('setFontSize', ISetFontSize),
        occurence.ZeroOrMore('scale', IScale),
        occurence.ZeroOrMore('translate', ITranslate),
        occurence.ZeroOrMore('rotate', IRotate),
        occurence.ZeroOrMore('skew', ISkew),
        occurence.ZeroOrMore('transform', ITransform),
        occurence.ZeroOrMore('lineMode', ILineMode),
        # Form Field Elements
        occurence.ZeroOrMore('barCode', form.IBarCode),
        occurence.ZeroOrMore('textField', form.ITextField),
        occurence.ZeroOrMore('buttonField', form.IButtonField),
        occurence.ZeroOrMore('selectField', form.ISelectField),
        # Charts
        occurence.ZeroOrMore('barChart', chart.IBarChart),
        occurence.ZeroOrMore('barChart3D', chart.IBarChart3D),
        occurence.ZeroOrMore('linePlot', chart.ILinePlot),
        occurence.ZeroOrMore('linePlot3D', chart.ILinePlot3D),
        occurence.ZeroOrMore('pieChart', chart.IPieChart),
        occurence.ZeroOrMore('pieChart3D', chart.IPieChart3D),
        occurence.ZeroOrMore('spiderChart', chart.ISpiderChart),
        # Misc
        occurence.ZeroOrMore('bookmark', IBookmark),
        occurence.ZeroOrMore('plugInGraphic', IPlugInGraphic),
    )


class Drawing(directive.RMLDirective):
    signature = IDrawing

    factories = {
        # State Management
        'saveState': SaveState,
        'restoreState': RestoreState,
        # Drawing Strings
        'drawString': DrawString,
        'drawRightString': DrawRightString,
        'drawCenteredString': DrawCenteredString,
        'drawCentredString': DrawCenteredString,
        'drawAlignedString': DrawAlignedString,
        # Drawing Operations
        'ellipse': Ellipse,
        'circle': Circle,
        'rect': Rectangle,
        'grid': Grid,
        'lines': Lines,
        'curves': Curves,
        'image': Image,
        'place': Place,
        'textAnnotation': TextAnnotation,
        'path': Path,
        # Form Field Elements
        'barCode': form.BarCode,
        'textField': form.TextField,
        'buttonField': form.ButtonField,
        'selectField': form.SelectField,
        # State Change Operations
        'fill': Fill,
        'stroke': Stroke,
        'setFont': SetFont,
        'setFontSize': SetFontSize,
        'scale': Scale,
        'translate': Translate,
        'rotate': Rotate,
        'skew': Skew,
        'transform': Transform,
        'lineMode': LineMode,
        # Charts
        'barChart': chart.BarChart,
        'barChart3D': chart.BarChart3D,
        'linePlot': chart.LinePlot,
        'linePlot3D': chart.LinePlot3D,
        'pieChart': chart.PieChart,
        'pieChart3D': chart.PieChart3D,
        'spiderChart': chart.SpiderChart,
        # Misc
        'bookmark': Bookmark,
        'plugInGraphic': PlugInGraphic,
    }


class IPageDrawing(IDrawing):
    """Draws directly on the content of one page's canvas. Every call of this
    directive creates a new page."""

    occurence.containing(
        # 'mergePage': IMergePage,
        *IDrawing.getTaggedValue('directives'))


class PageDrawing(Drawing):
    signature = IDrawing

    factories = Drawing.factories.copy()
    factories.update({
        'mergePage': page.MergePage
    })

    def process(self):
        super(Drawing, self).process()
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        canvas.showPage()


class IPageInfo(interfaces.IRMLDirectiveSignature):
    """Set's up page-global settings."""

    pageSize = attr.PageSize(
        title='Page Size',
        description=('The page size of all pages within this document.'),
        required=True)


class PageInfo(CanvasRMLDirective):
    signature = IPageInfo
    callable = 'setPageSize'
    attrMapping = {'pageSize': 'size'}



# Hook up our custom paragraph parser.
import importlib.metadata

from reportlab.lib.styles import getSampleStyleSheet

import z3c.rml.paraparser
import z3c.rml.rlfix  # noqa: F401 imported but unused


__version__ = importlib.metadata.version("z3c.rml")

del importlib.metadata


SampleStyleSheet = getSampleStyleSheet()



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Style Related Element Processing
"""
import reportlab.platypus.flowables
import reportlab.rl_config
import zope.interface
from reportlab.rl_config import overlapAttachedSpace

from z3c.rml import interfaces


# Fix problem with reportlab 3.1.44
class KeepInFrame(reportlab.platypus.flowables.KeepInFrame):
    pass

    def __init__(self, maxWidth, maxHeight, content=[], mergeSpace=1,
                 mode='shrink', name='', hAlign='LEFT', vAlign='BOTTOM',
                 fakeWidth=None):

        self.name = name
        self.maxWidth = maxWidth
        self.maxHeight = maxHeight
        self.mode = mode
        assert mode in ('error', 'overflow', 'shrink', 'truncate'), \
            f'{self.identity()} invalid mode value {mode}'
        # This is an unnecessary check, since wrap() handles None just fine!
        # assert maxHeight>=0,  \
        #       '%s invalid maxHeight value %s' % (self.identity(),maxHeight)
        if mergeSpace is None:
            mergeSpace = overlapAttachedSpace
        self.mergespace = mergeSpace
        self._content = content or []
        self.vAlign = vAlign
        self.hAlign = hAlign
        self.fakeWidth = fakeWidth


class BaseFlowable(reportlab.platypus.flowables.Flowable):
    def __init__(self, *args, **kw):
        reportlab.platypus.flowables.Flowable.__init__(self)
        self.args = args
        self.kw = kw

    def wrap(self, *args):
        return (0, 0)

    def draw(self):
        pass


class Illustration(reportlab.platypus.flowables.Flowable):
    def __init__(self, processor, width, height):
        self.processor = processor
        self.width = width
        self.height = height

    def wrap(self, *args):
        return (self.width, self.height)

    def draw(self):
        # Import here to avoid recursive imports
        from z3c.rml import canvas
        self.canv.saveState()
        drawing = canvas.Drawing(
            self.processor.element, self.processor)
        zope.interface.alsoProvides(drawing, interfaces.ICanvasManager)
        drawing.canvas = self.canv
        drawing.process()
        self.canv.restoreState()


class BookmarkPage(BaseFlowable):
    def draw(self):
        self.canv.bookmarkPage(*self.args, **self.kw)


class Bookmark(BaseFlowable):
    def draw(self):
        self.canv.bookmarkHorizontal(*self.args, **self.kw)


class OutlineAdd(BaseFlowable):
    def draw(self):
        if self.kw.get('key', None) is None:
            self.kw['key'] = str(hash(self))
        self.canv.bookmarkPage(self.kw['key'])
        self.canv.addOutlineEntry(**self.kw)


class Link(reportlab.platypus.flowables._Container,
           reportlab.platypus.flowables.Flowable):

    def __init__(self, content, **args):
        self._content = content
        self.args = args

    def wrap(self, availWidth, availHeight):
        self.width, self.height = reportlab.platypus.flowables._listWrapOn(
            self._content, availWidth, self.canv)
        return self.width, self.height

    def drawOn(self, canv, x, y, _sW=0, scale=1.0, content=None, aW=None):
        '''we simulate being added to a frame'''
        pS = 0
        if aW is None:
            aW = self.width
        aW = scale * (aW + _sW)
        if content is None:
            content = self._content
        y += self.height * scale

        startX = x
        startY = y
        totalWidth = 0
        totalHeight = 0

        for c in content:
            w, h = c.wrapOn(canv, aW, 0xfffffff)
            if w < reportlab.rl_config._FUZZ or h < reportlab.rl_config._FUZZ:
                continue
            if c is not content[0]:
                h += max(c.getSpaceBefore() - pS, 0)
            y -= h
            c.drawOn(canv, x, y, _sW=aW - w)
            if c is not content[-1]:
                pS = c.getSpaceAfter()
                y -= pS
            totalWidth += w
            totalHeight += h
        rectangle = [startX, startY - totalHeight,
                     startX + totalWidth, startY]
        if 'url' in self.args:
            canv.linkURL(rect=rectangle, **self.args)
        else:
            canv.linkAbsolute('', Rect=rectangle, **self.args)



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Style Related Element Processing
"""

import zope.interface
from reportlab import platypus

from z3c.rml import attr
from z3c.rml import canvas
from z3c.rml import directive
from z3c.rml import flowable
from z3c.rml import interfaces
from z3c.rml import occurence
from z3c.rml import page
from z3c.rml import stylesheet  # noqa: F401 imported but unused


class IStory(flowable.IFlow):
    """The story of the PDF file."""
    occurence.containing(
        occurence.ZeroOrMore('pto', flowable.IPTO),
        *flowable.IFlow.getTaggedValue('directives'))

    firstPageTemplate = attr.Text(
        title='First Page Template',
        description='The first page template to be used.',
        default=None,
        required=False)


class Story(flowable.Flow):
    signature = IStory

    factories = dict(
        pto=flowable.PTO,
        **flowable.Flow.factories
    )

    def process(self):
        self.parent.flowables = super().process()
        self.parent.doc._firstPageTemplateIndex = self.getFirstPTIndex()

    def getFirstPTIndex(self):
        args = dict(self.getAttributeValues(select=('firstPageTemplate',)))
        fpt = args.pop('firstPageTemplate', None)
        if fpt is None:
            return 0
        for idx, pageTemplate in enumerate(self.parent.doc.pageTemplates):
            if pageTemplate.id == fpt:
                return idx
        raise ValueError('%r is not a correct page template id.' % fpt)


class IFrame(interfaces.IRMLDirectiveSignature):
    """A frame on a page."""

    x1 = attr.Measurement(
        title='X-Position',
        description='The X-Position of the lower-left corner of the frame.',
        allowPercentage=True,
        required=True)

    y1 = attr.Measurement(
        title='Y-Position',
        description='The Y-Position of the lower-left corner of the frame.',
        allowPercentage=True,
        required=True)

    width = attr.Measurement(
        title='Width',
        description='The width of the frame.',
        allowPercentage=True,
        required=True)

    height = attr.Measurement(
        title='Height',
        description='The height of the frame.',
        allowPercentage=True,
        required=True)

    id = attr.Text(
        title='Id',
        description='The id of the frame.',
        required=False)

    leftPadding = attr.Measurement(
        title='Left Padding',
        description='The left padding of the frame.',
        default=0,
        required=False)

    rightPadding = attr.Measurement(
        title='Right Padding',
        description='The right padding of the frame.',
        default=0,
        required=False)

    topPadding = attr.Measurement(
        title='Top Padding',
        description='The top padding of the frame.',
        default=0,
        required=False)

    bottomPadding = attr.Measurement(
        title='Bottom Padding',
        description='The bottom padding of the frame.',
        default=0,
        required=False)

    showBoundary = attr.Boolean(
        title='Show Boundary',
        description='A flag to show the boundary of the frame.',
        required=False)


class Frame(directive.RMLDirective):
    signature = IFrame

    def process(self):
        # get the page size
        size = self.parent.pt.pagesize
        if size is None:
            size = self.parent.parent.parent.doc.pagesize
        # Get the arguments
        args = dict(self.getAttributeValues())
        # Deal with percentages
        for name, dir in (('x1', 0), ('y1', 1), ('width', 0), ('height', 1)):
            if (isinstance(args[name], str) and
                    args[name].endswith('%')):
                args[name] = float(args[name][:-1]) / 100 * size[dir]
        frame = platypus.Frame(**args)
        self.parent.frames.append(frame)


class IPageGraphics(canvas.IDrawing):
    """Define the page graphics for the page template."""


@zope.interface.implementer(interfaces.ICanvasManager)
class PageGraphics(directive.RMLDirective):
    signature = IPageGraphics

    def process(self):
        onPage = self.parent.pt.onPage

        def drawOnCanvas(canv, doc):
            onPage(canv, doc)
            canv.saveState()
            self.canvas = canv
            drawing = canvas.Drawing(self.element, self)
            drawing.process()
            canv.restoreState()

        self.parent.pt.onPage = drawOnCanvas


class Header(PageGraphics):

    def process(self):
        onPage = self.parent.pt.onPage

        def drawOnCanvas(canv, doc):
            onPage(canv, doc)
            canv.saveState()
            self.canvas = canv
            place = canvas.Place(self.element, self)
            place.process()
            canv.restoreState()

        self.parent.pt.onPage = drawOnCanvas


class Footer(Header):
    pass


class IPageTemplate(interfaces.IRMLDirectiveSignature):
    """Define a page template."""
    occurence.containing(
        occurence.OneOrMore('frame', IFrame),
        occurence.ZeroOrOne('pageGraphics', IPageGraphics),
        occurence.ZeroOrOne('mergePage', page.IMergePage),
    )

    id = attr.Text(
        title='Id',
        description='The id of the template.',
        required=True)

    pagesize = attr.PageSize(
        title='Page Size',
        description='The Page Size.',
        required=False)

    autoNextTemplate = attr.Text(
        title='Auto Next Page Template',
        description='The page template to use automatically for the next'
                    ' page.',
        required=False)


class PageTemplate(directive.RMLDirective):
    signature = IPageTemplate
    attrMapping = {'autoNextTemplate': 'autoNextPageTemplate'}
    factories = {
        'frame': Frame,
        'pageGraphics': PageGraphics,
        'mergePage': page.MergePageInPageTemplate,
        'header': Header,
        'footer': Footer,
    }

    def process(self):
        args = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        pagesize = args.pop('pagesize', None)

        self.frames = []
        self.pt = platypus.PageTemplate(**args)

        self.processSubDirectives()
        self.pt.frames = self.frames

        if pagesize:
            self.pt.pagesize = pagesize

        self.parent.parent.doc.addPageTemplates(self.pt)


class ITemplate(interfaces.IRMLDirectiveSignature):
    """Define a page template."""
    occurence.containing(
        occurence.OneOrMore('pageTemplate', IPageTemplate),
    )

    pagesize = attr.PageSize(
        title='Page Size',
        description='The Page Size.',
        required=False)

    rotation = attr.Integer(
        title='Rotation',
        description='The rotation of the page in multiples of 90 degrees.',
        required=False)

    leftMargin = attr.Measurement(
        title='Left Margin',
        description='The left margin of the template.',
        default=0,
        required=False)

    rightMargin = attr.Measurement(
        title='Right Margin',
        description='The right margin of the template.',
        default=0,
        required=False)

    topMargin = attr.Measurement(
        title='Top Margin',
        description='The top margin of the template.',
        default=0,
        required=False)

    bottomMargin = attr.Measurement(
        title='Bottom Margin',
        description='The bottom margin of the template.',
        default=0,
        required=False)

    showBoundary = attr.Boolean(
        title='Show Boundary',
        description='A flag to show the boundary of the template.',
        required=False)

    allowSplitting = attr.Boolean(
        title='Allow Splitting',
        description='A flag to allow splitting over multiple templates.',
        required=False)

    title = attr.Text(
        title='Title',
        description='The title of the PDF document.',
        required=False)

    author = attr.Text(
        title='Author',
        description='The author of the PDF document.',
        required=False)


class Template(directive.RMLDirective):
    signature = ITemplate
    factories = {
        'pageTemplate': PageTemplate,
    }

    def process(self):
        args = self.getAttributeValues()
        args += self.parent.getAttributeValues(
            select=('debug', 'compression', 'invariant'),
            attrMapping={'debug': '_debug', 'compression': 'pageCompression'})
        args += (('cropMarks', self.parent.cropMarks),)

        self.parent.doc = platypus.BaseDocTemplate(
            self.parent.outputFile, **dict(args))
        self.processSubDirectives()



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""RML Reference Generator
"""
import os
import re
from xml.sax import saxutils

import pygments.token
import zope.schema
import zope.schema.interfaces
from lxml import etree
from pygments.lexers import XmlLexer

from z3c.rml import __version__
from z3c.rml import attr
from z3c.rml import document
from z3c.rml import interfaces
from z3c.rml import pagetemplate


INPUT_URL = ('https://github.com/zopefoundation/z3c.rml/blob/master/src/z3c/'
             'rml/tests/input/%s')
EXPECTED_URL = (
    'https://github.com/zopefoundation/z3c.rml/blob/master/src/z3c/'
    'rml/tests/expected/%s?raw=true')

EXAMPLES_DIRECTORY = os.path.join(os.path.dirname(__file__), 'tests', 'input')
IGNORE_ATTRIBUTES = ('RMLAttribute', 'BaseChoice')
CONTENT_FIELD_TYPES = (
    attr.TextNode, attr.TextNodeSequence, attr.TextNodeGrid,
    attr.RawXMLContent, attr.XMLContent)
STYLES_FORMATTING = {
    pygments.token.Name.Tag: ('<font textColor="red">', '</font>'),
    pygments.token.Literal.String: ('<font textColor="blue">', '</font>'),
}
EXAMPLE_NS = 'http://namespaces.zope.org/rml/doc'
EXAMPLE_ATTR_NAME = '{%s}example' % EXAMPLE_NS


def dedent(rml):
    spaces = re.findall('\n( *)<', rml)
    if not spaces:
        return rml
    least = min([len(s) for s in spaces if s != ''])
    return rml.replace('\n' + ' ' * least, '\n')


def enforceColumns(rml, columns=80):
    result = []
    for line in rml.split('\n'):
        if len(line) <= columns:
            result.append(line)
            continue
        # Determine the indentation for all other lines
        lineStart = re.findall('^( *<[a-zA-Z0-9]+ )', line)
        lineIndent = 0
        if lineStart:
            lineIndent = len(lineStart[0])
        # Create lines having at most the specified number of columns
        while len(line) > columns:
            end = line[:columns].rfind(' ')
            result.append(line[:end])
            line = ' ' * lineIndent + line[end + 1:]
        result.append(line)

    return '\n'.join(result)


def highlightRML(rml):
    lexer = XmlLexer()
    styledRml = ''
    for ttype, token in lexer.get_tokens(rml):
        start, end = STYLES_FORMATTING.get(ttype, ('', ''))
        styledRml += start + saxutils.escape(token) + end
    return styledRml


def removeDocAttributes(elem):
    for name in elem.attrib.keys():
        if name.startswith('{' + EXAMPLE_NS + '}'):
            del elem.attrib[name]
    for child in elem.getchildren():
        removeDocAttributes(child)


def getAttributeTypes():
    types = []
    candidates = sorted(attr.__dict__.items(), key=lambda e: e[0])
    for name, candidate in candidates:
        if not (isinstance(candidate, type) and
                zope.schema.interfaces.IField.implementedBy(candidate) and
                name not in IGNORE_ATTRIBUTES):
            continue
        types.append({
            'name': name,
            'description': candidate.__doc__
        })
    return types


def formatField(field):
    return field.__class__.__name__


def formatChoice(field):
    choices = ', '.join([repr(choice) for choice in field.choices.keys()])
    return f'{field.__class__.__name__} of ({choices})'


def formatSequence(field):
    vtFormatter = typeFormatters.get(field.value_type.__class__, formatField)
    return '{} of {}'.format(
        field.__class__.__name__, vtFormatter(field.value_type))


def formatGrid(field):
    vtFormatter = typeFormatters.get(field.value_type.__class__, formatField)
    return '%s with %i cols of %s' % (
        field.__class__.__name__, field.columns, vtFormatter(field.value_type))


def formatCombination(field):
    vts = [typeFormatters.get(vt.__class__, formatField)(vt)
           for vt in field.value_types]
    return '{} of {}'.format(field.__class__.__name__, ', '.join(vts))


typeFormatters = {
    attr.Choice: formatChoice,
    attr.Sequence: formatSequence,
    attr.Combination: formatCombination,
    attr.TextNodeSequence: formatSequence,
    attr.TextNodeGrid: formatGrid}


def processSignature(name, signature, queue, examples, directives=None):
    if directives is None:
        directives = {}
    # Process this directive
    if signature not in directives:
        info = {'name': name, 'description': signature.getDoc(),
                'id': str(hash(signature)), 'deprecated': False}
        # If directive is deprecated, then add some info
        if interfaces.IDeprecatedDirective.providedBy(signature):
            info['deprecated'] = True
            info['reason'] = signature.getTaggedValue('deprecatedReason')
        attrs = []
        content = None
        for fname, field in zope.schema.getFieldsInOrder(signature):
            # Handle the case, where the field describes the content
            typeFormatter = typeFormatters.get(field.__class__, formatField)
            fieldInfo = {
                'name': fname,
                'type': typeFormatter(field),
                'title': field.title,
                'description': field.description,
                'required': field.required,
                'deprecated': False,
            }
            if field.__class__ in CONTENT_FIELD_TYPES:
                content = fieldInfo
            else:
                attrs.append(fieldInfo)

            # Add a separate entry for the deprecated field
            if interfaces.IDeprecated.providedBy(field):
                deprFieldInfo = fieldInfo.copy()
                deprFieldInfo['deprecated'] = True
                deprFieldInfo['name'] = field.deprecatedName
                deprFieldInfo['reason'] = field.deprecatedReason
                attrs.append(deprFieldInfo)

        info['attributes'] = attrs
        info['content'] = content
        # Examples can be either gotten by interface path or tag name
        ifacePath = signature.__module__ + '.' + signature.__name__
        if ifacePath in examples:
            info['examples'] = examples[ifacePath]
        else:
            info['examples'] = examples.get(name, None)

        subs = []
        for occurence in signature.queryTaggedValue('directives', ()):
            subs.append({
                'name': occurence.tag,
                'occurence': occurence.__class__.__name__,
                'deprecated': interfaces.IDeprecatedDirective.providedBy(
                    occurence.signature),
                'id': str(hash(occurence.signature))
            })
        info['sub-directives'] = subs
        directives[signature] = info

        # Process Children
        for occurence in signature.queryTaggedValue('directives', ()):
            queue.append((occurence.tag, occurence.signature))


def extractExamples(directory):
    examples = {}
    for filename in os.listdir(directory):
        if not filename.endswith('.rml'):
            continue
        rmlFile = open(os.path.join(directory, filename), 'rb')
        root = etree.parse(rmlFile).getroot()
        elements = root.xpath('//@doc:example/parent::*',
                              namespaces={'doc': EXAMPLE_NS})
        # Phase 1: Collect all elements
        for elem in elements:
            demoTag = elem.get(EXAMPLE_ATTR_NAME) or elem.tag
            elemExamples = examples.setdefault(demoTag, [])
            elemExamples.append({
                'filename': filename,
                'line': elem.sourceline,
                'element': elem,
                'rmlurl': INPUT_URL % filename,
                'pdfurl': EXPECTED_URL % (filename[:-4] + '.pdf')
            })
        # Phase 2: Render all elements
        removeDocAttributes(root)
        for dirExamples in examples.values():
            for example in dirExamples:
                xml = etree.tounicode(example['element']).strip()
                xml = re.sub(
                    ' ?xmlns:doc="http://namespaces.zope.org/rml/doc"', '', xml
                )
                xml = dedent(xml)
                xml = enforceColumns(xml, 80)
                xml = highlightRML(xml)
                example['code'] = xml

        rmlFile.close()

    return examples


def main(outPath=None):
    examples = extractExamples(EXAMPLES_DIRECTORY)

    template = pagetemplate.RMLPageTemplateFile('reference.pt')

    directives = {}
    queue = [('document', document.IDocument)]
    while queue:
        tag, sig = queue.pop()
        processSignature(tag, sig, queue, examples, directives)
    directives = sorted(directives.values(), key=lambda d: d['name'])

    pdf = template(
        version=__version__,
        types=getAttributeTypes(),
        directives=directives)
    file_ = open(outPath or 'rml-reference.pdf', 'wb')
    file_.write(pdf)
    file_.close()



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Number to Words converter.
"""
__docformat__ = "reStructuredText"

import math


def toOrdinal(num):
    str_num = str(num)
    str_num += "tsnrhtdd"[(num / 10 % 10 != 1) * (num % 10 < 4) * num % 10::4]
    return str_num


class Number2Words:

    units = [
        "zero", "one", "two", "three", "four", "five", "six", "seven", "eight",
        "nine", "ten", "eleven", "twelve", "thirteen", "fourteen", "fifteen",
        "sixteen", "seventeen", "eighteen", "nineteen",
    ]

    units_ordinal = [
        "zeroth", "first", "second", "third", "fourth", "fifth", "sixth",
        "seventh", "eighth", "ninth", "tenth", "eleventh", "twelfth",
        "thirteenth", "fourteenth", "fifteenth", "sixteenth", "seventeenth",
        "eighteenth", "nineteenth",
    ]

    tens = ["", "", "twenty", "thirty", "forty", "fifty", "sixty", "seventy",
            "eighty", "ninety"]

    tens_ordinal = ["", "", "twentieth", "thirtieth", "fortieth", "fiftieth",
                    "sixtieth", "seventieth", "eightieth", "ninetieth"]

    scales = ["", "", "hundred", "thousand", "", "", "million", "", "",
              "billion", "", "", "trillion"]

    def __call__(self, num, as_list=False, ordinal=False):
        num_str = str(num)
        num_left = num_str
        words = []
        while num_left:
            if int(num_left) < 20:
                if int(num_left) != 0 or not len(words):
                    words.append(self.units[int(num_left)])
                break
            if int(num_left) < 100:
                words.append(self.tens[int(num_left[0])])
                num_left = num_left[1:]
            elif int(num_left) < 1000:
                scale = self.scales[len(num_left) - 1]
                digit = int(num_left[0])
                if digit != 0:
                    words.append(self.units[digit])
                words.append(scale)
                num_left = num_left[1:]
            else:
                mag = int(math.log10(int(num_left)) / 3) * 3
                words += self(num_left[:-mag], True) + [self.scales[mag]]
                num_left = num_left[-mag:]

        if ordinal:
            if words[-1] in self.units:
                words[-1] = self.units_ordinal[self.units.index(words[-1])]
            elif words[-1] in self.tens:
                words[-1] = self.tens_ordinal[self.tens.index(words[-1])]
            elif words[-1] in self.scales:
                words[-1] += 'th'
        words = [w.title() for w in words if w != '']
        if as_list:
            return words
        # two digit numbers require dashes
        if len(words) > 1:
            result = []
            tens = {w.title() for w in self.tens}
            units = {w.title() for w in self.units + self.units_ordinal}
            while words:
                cur = words.pop(0)
                if cur in tens and words[0] in units:
                    ones = words.pop(0)
                    result.append(f'{cur}-{ones}')
                else:
                    result.append(cur)
            words = result

        return ' '.join(words)


num2words = Number2Words()



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTLAR PURPOSE.
#
##############################################################################
"""Testing all XML Locale functionality.
"""
import base64
import logging
import os
import subprocess
import sys
import tempfile
import unittest

from PIL import Image
from PIL import ImageChops

import z3c.rml.tests
from z3c.rml import rml2pdf


LOG_FILE = os.path.join(os.path.dirname(__file__), 'render.log')


def gs_command(path):
    return ('gs', '-q', '-sNOPAUSE', '-sDEVICE=png256',
            '-sOutputFile=%s[Page-%%d].png' % path[:-4],
            path, '-c', 'quit')


class RMLRenderingTestCase(unittest.TestCase):

    def __init__(self, inPath, outPath):
        self._inPath = inPath
        self._outPath = outPath
        unittest.TestCase.__init__(self)

    def setUp(self):
        import z3c.rml.tests.module
        sys.modules['module'] = z3c.rml.tests.module
        sys.modules['mymodule'] = z3c.rml.tests.module

    def tearDown(self):
        del sys.modules['module']
        del sys.modules['mymodule']

        from z3c.rml.document import LOGGER_NAME
        for handler in logging.getLogger(LOGGER_NAME).handlers:
            if handler.baseFilename == LOG_FILE:
                handler.close()

        if os.path.exists(LOG_FILE):
            os.remove(LOG_FILE)

    def runTest(self):
        rml2pdf.go(self._inPath, self._outPath)


class ComparePDFTestCase(unittest.TestCase):

    level = 2

    def __init__(self, basePath, testPath):
        self._basePath = basePath
        self._testPath = testPath
        unittest.TestCase.__init__(self)

    def assertSameImage(self, baseImage, testImage):
        base_file = open(baseImage, 'rb')
        test_file = open(testImage, 'rb')
        base_image = Image.open(base_file)
        base = base_image.getdata()
        test_image = Image.open(test_file)
        test = test_image.getdata()

        has_diff = True
        for i in range(len(base)):
            # Stop at first difference
            if (base[i] - test[i]) != 0:
                break
        else:
            has_diff = False

        if has_diff and 'SHOW_IMAGE_DIFF' in os.environ:
            test_image.show()
            base_image.show()
            differ = ImageChops.subtract(test_image, base_image)
            differ.show()
            differ2 = ImageChops.subtract(base_image, test_image)
            differ2.show()
            # output the result as base64 for travis debugging
            if 'SHOW_DIFF_DEBUG' in os.environ:
                print(base64.b64encode(differ.tobytes()))
                print(base64.b64encode(differ2.tobytes()))

        if has_diff and 'SHOW_DIFF_DEBUG' in os.environ:
            print()
            print(os.system("gs --version"))

            base_file.seek(0)
            print(baseImage)
            print(base64.b64encode(base_file.read()))
            print(self._basePath)
            with open(self._basePath, 'rb') as base_pdf:
                print(base64.b64encode(base_pdf.read()))

            test_file.seek(0)
            print(testImage)
            print(base64.b64encode(test_file.read()))
            print(self._testPath)
            with open(self._testPath, 'rb') as test_pdf:
                print(base64.b64encode(test_pdf.read()))

        base_file.close()
        test_file.close()

        if has_diff:
            self.fail(
                'Image is not the same: %s' % os.path.basename(baseImage))

    def runTest(self):
        # Convert the base PDF to image(s)
        status = subprocess.Popen(gs_command(self._basePath)).wait()
        if status:
            return
        # Convert the test PDF to image(s)
        status = subprocess.Popen(gs_command(self._testPath)).wait()
        if status:
            return
        # Go through all pages and ensure their equality
        n = 1
        while True:
            baseImage = self._basePath[:-4] + '[Page-%i].png' % n
            testImage = self._testPath[:-4] + '[Page-%i].png' % n
            if os.path.exists(baseImage) and os.path.exists(testImage):
                self.assertSameImage(baseImage, testImage)
            else:
                break
            n += 1


class CompareFileTestCase(unittest.TestCase):

    def __init__(self, testPath, contains):
        self._testPath = testPath
        self._contains = contains
        unittest.TestCase.__init__(self)

    def runTest(self):
        with open(self._testPath, 'rb') as io:
            contents = io.read()

            if self._contains not in contents:
                self.fail(
                    'PDF file does not contain: %s' % self._contains
                )


def test_suite():
    if False:
        # Debug info
        import pkg_resources
        print("Environment\n===========")
        print("reportlab:")
        print(pkg_resources.get_distribution('reportlab').version)
        print("ghostscript:")
        os.system("gs --version")

    suite = unittest.TestSuite()
    here = os.path.dirname(z3c.rml.tests.__file__)
    inputDir = os.path.join(here, 'input')
    outputDir = tempfile.mkdtemp('z3c.rml-output')
    print(f'output dir: {outputDir}')
    expectDir = os.path.join(here, 'expected')
    for filename in os.listdir(inputDir):
        if not filename.endswith(".rml"):
            continue

        inPath = os.path.join(inputDir, filename)
        outPath = os.path.join(outputDir, filename[:-4] + '.pdf')
        expectPath = os.path.join(expectDir, filename[:-4] + '.pdf')

        # ** Test RML to PDF rendering **
        # Create new type, so that we can get test matching
        TestCase = type(filename[:-4], (RMLRenderingTestCase,), {})
        case = TestCase(inPath, outPath)
        suite.addTest(case)
        # ** Test PDF rendering correctness **
        TestCase = type('compare-' + filename[:-4], (ComparePDFTestCase,), {})
        case = TestCase(expectPath, outPath)
        suite.addTest(case)

        if filename == 'printScaling.rml':
            TestCase = type('compare-file-' + filename[:-4],
                            (CompareFileTestCase,), {})
            case = TestCase(outPath, b'/PrintScaling /None')
            suite.addTest(case)

    return suite



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""``ul``, ``ol``, and ``li`` directives.
"""
__docformat__ = "reStructuredText"
import copy

import reportlab.lib.styles
import reportlab.platypus
import zope.schema
from reportlab.platypus import flowables  # noqa: F401 imported but unused

from z3c.rml import attr
from z3c.rml import directive
from z3c.rml import flowable
from z3c.rml import interfaces
from z3c.rml import occurence
from z3c.rml import stylesheet


class IListItem(stylesheet.IMinimalListStyle, flowable.IFlow):
    """A list item in an ordered or unordered list."""

    style = attr.Style(
        title='Style',
        description='The list style that is applied to the list.',
        required=False)


class ListItem(flowable.Flow):
    signature = IListItem
    klass = reportlab.platypus.ListItem
    attrMapping = {}

    styleAttributes = zope.schema.getFieldNames(stylesheet.IMinimalListStyle)

    def processStyle(self, style):
        attrs = self.getAttributeValues(select=self.styleAttributes)
        if attrs or not hasattr(style, 'value'):
            style = copy.deepcopy(style)
            # Sigh, this is needed since unordered list items expect the value.
            style.value = style.start
            for name, value in attrs:
                setattr(style, name, value)
        return style

    def process(self):
        self.processSubDirectives()
        args = dict(self.getAttributeValues(ignore=self.styleAttributes))
        if 'style' not in args:
            args['style'] = self.parent.baseStyle
        args['style'] = self.processStyle(args['style'])
        li = self.klass(self.flow, **args)
        self.parent.flow.append(li)


class IOrderedListItem(IListItem):
    """An ordered list item."""

    value = attr.Integer(
        title='Bullet Value',
        description='The counter value.',
        required=False)


class OrderedListItem(ListItem):
    signature = IOrderedListItem


class IUnorderedListItem(IListItem):
    """An ordered list item."""

    value = attr.Combination(
        title='Bullet Value',
        description='The type of bullet character.',
        value_types=(
            attr.Choice(choices=interfaces.UNORDERED_BULLET_VALUES),
            attr.Text(max_length=1)
        ),
        required=False)


class UnorderedListItem(ListItem):
    signature = IUnorderedListItem

    styleAttributes = ListItem.styleAttributes + ['value']


class IListBase(stylesheet.IBaseListStyle):

    style = attr.Style(
        title='Style',
        description='The list style that is applied to the list.',
        required=False)


class ListBase(directive.RMLDirective):
    klass = reportlab.platypus.ListFlowable
    factories = {'li': ListItem}
    attrMapping = {}

    styleAttributes = zope.schema.getFieldNames(stylesheet.IBaseListStyle)

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.flow = []

    def processStyle(self, style):
        attrs = self.getAttributeValues(
            select=self.styleAttributes, attrMapping=self.attrMapping)
        if attrs:
            style = copy.deepcopy(style)
            for name, value in attrs:
                setattr(style, name, value)
        return style

    def process(self):
        args = dict(self.getAttributeValues(
            ignore=self.styleAttributes, attrMapping=self.attrMapping))
        if 'style' not in args:
            args['style'] = reportlab.lib.styles.ListStyle('List')
        args['style'] = self.baseStyle = self.processStyle(args['style'])
        self.processSubDirectives()
        li = self.klass(self.flow, **args)
        self.parent.flow.append(li)


class IOrderedList(IListBase):
    """An ordered list."""
    occurence.containing(
        occurence.ZeroOrMore('li', IOrderedListItem),
    )

    bulletType = attr.Choice(
        title='Bullet Type',
        description='The type of bullet formatting.',
        choices=interfaces.ORDERED_LIST_TYPES,
        doLower=False,
        required=False)


class OrderedList(ListBase):
    signature = IOrderedList
    factories = {'li': OrderedListItem}

    styleAttributes = ListBase.styleAttributes + ['bulletType']


class IUnorderedList(IListBase):
    """And unordered list."""
    occurence.containing(
        occurence.ZeroOrMore('li', IUnorderedListItem),
    )

    value = attr.Choice(
        title='Bullet Value',
        description='The type of bullet character.',
        choices=interfaces.UNORDERED_BULLET_VALUES,
        default='circle',
        required=False)


class UnorderedList(ListBase):
    signature = IUnorderedList
    attrMapping = {'value': 'start'}
    factories = {'li': UnorderedListItem}

    def getAttributeValues(self, *args, **kw):
        res = super().getAttributeValues(*args, **kw)
        res.append(('bulletType', 'bullet'))
        return res


flowable.Flow.factories['ol'] = OrderedList
flowable.IFlow.setTaggedValue(
    'directives',
    flowable.IFlow.getTaggedValue('directives') +
    (occurence.ZeroOrMore('ol', IOrderedList),)
)

flowable.Flow.factories['ul'] = UnorderedList
flowable.IFlow.setTaggedValue(
    'directives',
    flowable.IFlow.getTaggedValue('directives') +
    (occurence.ZeroOrMore('ul', IUnorderedList),)
)



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""RML to PDF Converter
"""
import argparse
import io
import os
import sys

import zope.interface
from lxml import etree

from z3c.rml import document
from z3c.rml import interfaces


zope.interface.moduleProvides(interfaces.IRML2PDF)


def parseString(xml, removeEncodingLine=True, filename=None):
    if isinstance(xml, str) and removeEncodingLine:
        # RML is a unicode string, but oftentimes documents declare their
        # encoding using <?xml ...>. Unfortuantely, I cannot tell lxml to
        # ignore that directive. Thus we remove it.
        if xml.startswith('<?xml'):
            xml = xml.split('\n', 1)[-1]
    root = etree.fromstring(xml)
    doc = document.Document(root)
    if filename:
        doc.filename = filename
    output = io.BytesIO()
    doc.process(output)
    output.seek(0)
    return output


def go(xmlInputName, outputFileName=None, outDir=None, dtdDir=None):
    if hasattr(xmlInputName, 'read'):
        # it is already a file-like object
        xmlFile = xmlInputName
        xmlInputName = 'input.pdf'
    else:
        with open(xmlInputName, 'rb') as xmlFile:
            return go(xmlFile, outputFileName, outDir, dtdDir)

    # If an output filename is specified, create an output file for it
    outputFile = None
    if outputFileName is not None:
        if hasattr(outputFileName, 'write'):
            # it is already a file-like object
            outputFile = outputFileName
            outputFileName = 'output.pdf'
        else:
            if outDir is not None:
                outputFileName = os.path.join(outDir, outputFileName)
            with open(outputFileName, 'wb') as outputFile:
                return go(xmlFile, outputFile, outDir, dtdDir)

    if dtdDir is not None:
        sys.stderr.write('The ``dtdDir`` option is not yet supported.\n')

    root = etree.parse(xmlFile).getroot()
    doc = document.Document(root)
    doc.filename = xmlInputName

    # Create a Reportlab canvas by processing the document
    doc.process(outputFile)


def main(args=None):
    if args is None:
        parser = argparse.ArgumentParser(
            prog='rml2pdf',
            description='Converts file in RML format into PDF file.',
            epilog='Copyright (c) 2007 Zope Foundation and Contributors.'
        )
        parser.add_argument('xmlInputName', help='RML file to be processed')
        parser.add_argument(
            'outputFileName',
            nargs='?',
            help='output PDF file name')
        parser.add_argument('outDir', nargs='?', help='output directory')
        parser.add_argument(
            'dtdDir',
            nargs='?',
            help='directory with XML DTD (not yet supported)')
        pargs = parser.parse_args()
        args = (
            pargs.xmlInputName,
            pargs.outputFileName,
            pargs.outDir,
            pargs.dtdDir)

    go(*args)


if __name__ == '__main__':
    canvas = go(sys.argv[1])



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Page Drawing Related Element Processing
"""
import io

from z3c.rml import attr
from z3c.rml import directive
from z3c.rml import interfaces


try:
    import pikepdf
    from pikepdf import Object  # noqa: F401 imported but unused
except ImportError:
    # We don't want to require pikepdf, if you do not want to use the features
    # in this module.
    pikepdf = None


def mergePage(layerPage, mainPage, pdf, name) -> None:
    contentsForName = pdf.copy_foreign(
        pikepdf.Page(layerPage).as_form_xobject()
    )
    newContents = b'q\n %s Do\nQ\n' % (name.encode())
    if not mainPage.Resources.get("/XObject"):
        mainPage.Resources["/XObject"] = pikepdf.Dictionary({})
    mainPage.Resources["/XObject"][name] = contentsForName
    # Use the MediaBox from the merged page
    mainPage.MediaBox = pikepdf.Array(layerPage.MediaBox)
    mainPage.contents_add(
        contents=pikepdf.Stream(pdf, newContents),
        prepend=True
    )


class MergePostProcessor:

    def __init__(self):
        self.operations = {}

    def process(self, inputFile1):
        input1 = pikepdf.open(inputFile1)
        count = 0
        for (num, page) in enumerate(input1.pages):
            if num in self.operations:
                for mergeFile, mergeNumber in self.operations[num]:
                    mergePdf = pikepdf.open(mergeFile)
                    toMerge = mergePdf.pages[mergeNumber]
                    name = f"/Fx{count}"
                    mergePage(toMerge, page, input1, name)

        outputFile = io.BytesIO()
        input1.save(outputFile)
        return outputFile


class IMergePage(interfaces.IRMLDirectiveSignature):
    """Merges an existing PDF Page into the one to be generated."""

    filename = attr.File(
        title='File',
        description=('Reference to the PDF file to extract the page from.'),
        required=True)

    page = attr.Integer(
        title='Page Number',
        description='The page number of the PDF file that is used to merge..',
        required=True)


class MergePage(directive.RMLDirective):
    signature = IMergePage

    def getProcessor(self):
        manager = attr.getManager(self, interfaces.IPostProcessorManager)
        procs = dict(manager.postProcessors)
        if 'MERGE' not in procs:
            proc = MergePostProcessor()
            manager.postProcessors.append(('MERGE', proc))
            return proc
        return procs['MERGE']

    def process(self):
        if pikepdf is None:
            raise Exception(
                'pikepdf is not installed, so this feature is not available.')
        inputFile, inPage = self.getAttributeValues(valuesOnly=True)
        manager = attr.getManager(self, interfaces.ICanvasManager)
        outPage = manager.canvas.getPageNumber() - 1

        proc = self.getProcessor()
        pageOperations = proc.operations.setdefault(outPage, [])
        pageOperations.append((inputFile, inPage))


class MergePageInPageTemplate(MergePage):

    def process(self):
        if pikepdf is None:
            raise Exception(
                'pikepdf is not installed, so this feature is not available.')
        inputFile, inPage = self.getAttributeValues(valuesOnly=True)

        onPage = self.parent.pt.onPage

        def drawOnCanvas(canvas, doc):
            onPage(canvas, doc)
            outPage = canvas.getPageNumber() - 1
            proc = self.getProcessor()
            pageOperations = proc.operations.setdefault(outPage, [])
            pageOperations.append((inputFile, inPage))

        self.parent.pt.onPage = drawOnCanvas



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Setup"""
import os

from setuptools import setup


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


TESTS_REQUIRE = [
    'mock',
    'Pillow',
    'coverage',
    'zope.pagetemplate',
    'zope.testrunner >= 6.4',
]


setup(
    name="z3c.rml",
    version='5.0',
    author="Stephan Richter and the Zope Community",
    author_email="zope-dev@zope.dev",
    description="An alternative implementation of RML",
    long_description=(
        read('README.rst')
        + '\n\n' +
        read('CHANGES.rst')
    ),
    license="ZPL-2.1",
    python_requires='>=3.9',
    keywords="rml reportlab pdf pagetemplate",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: Implementation :: CPython',
        'Natural Language :: English',
        'Operating System :: OS Independent',
    ],
    url='https://github.com/zopefoundation/z3c.rml',
    extras_require=dict(
        test=TESTS_REQUIRE,
        pagetemplate=[
            'zope.pagetemplate']
    ),
    install_requires=[
        'Pygments',
        'backports.tempfile',
        'lxml',
        'pikepdf>=3.0',
        'reportlab>=3.5.0',
        'rlPyCairo',
        'setuptools',
        'svglib',
        'zope.interface',
        'zope.schema',
    ],
    entry_points={
        'console_scripts': [
            'rml2pdf = z3c.rml.rml2pdf:main',
            'dtd = z3c.rml.dtd:main',
            'reference = z3c.rml.reference:main'],
    },
    include_package_data=True,
    zip_safe=False,
)



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTLAR PURPOSE.
#
##############################################################################
"""Test generating DTD
"""
import io
import os
import unittest

import lxml.etree

from z3c.rml import dtd


class DTDUnitTestCase(unittest.TestCase):

    def runTest(self):
        text = dtd.generate()
        et = lxml.etree.DTD(io.StringIO(text))
        self.assertIsNotNone(next(et.iterelements()))


class UpdateDTDTestCase(unittest.TestCase):

    level = 2

    def runTest(self):
        path = os.path.join(os.path.dirname(dtd.__file__), 'rml.dtd')
        # Write the file.
        with open(path, 'w') as file:
            file.write(dtd.generate())
        # Ensure we produced a processable DTD.
        try:
            lxml.etree.DTD(path)
        except BaseException:
            raise
